/**
 * @file
 * @copyright Swissbit Germany AG.
 * @author Kosma Moczek <kmoczek@fastlogic.pl>
 * @author Johannes Rath <johannes.rath@swissbit.com>
 */

/*
#define PRINT_ASSD
#define PRINT_VERBOSE
#define DO_CONFIG
#define CHECK_STACK
*/

#include <common.h>
#include <command.h>
#include <console.h>
#include <mmc.h>
#include <net.h>
#include <../net/bootp.h>
#include <../net/dns.h>
#include <../drivers/mmc/mmc_private.h>
#include <usb.h>
#include <blk.h>
#include <u-boot/sha256.h>
#include <memalign.h>

/* Driver model includes to get USB devices                                     */
#include <dm.h>
#include <dm/uclass-internal.h>


#define DP_CMD_ATTEMPTS 10
#define DP_ID_LENGTH 16
#define DP_PASSWORD_LENGTH 32
#define DP_PASSWORD_MAX_LENGTH 32
#define DP_CHALLENGE_LENGTH 32
#define DP_HASH_LENGTH 32
#define DP_AUTHENTICITY_LENGTH 32

typedef struct _STATUS_RESPONSE  {
	uint8_t mode;
	uint8_t state;
	int8_t counter;
	int8_t so_counter;
	uint32_t reset_counter;
	uint8_t rfu[2];
	uint32_t cdrom;
	uint8_t flags;
	uint16_t status;
} __attribute__((packed)) STATUS_RESPONSE;

typedef struct _STATUS_NVRAMRESPONSE {
	uint32_t hiddenSectors;
	uint8_t accessRandomRights;
	uint8_t accessCyclicRights;
	uint32_t randomSectors;
	uint32_t cyclicSectors;
	uint32_t nextSector;
	uint8_t wrap;
	uint16_t status;
} __attribute__((packed)) STATUS_NVRAMRESPONSE;

#ifdef CHECK_STACK
uint32_t top_of_stack;

void SizeOfStack(void)
{
	int x = 0;
	void *p = malloc(100);

#if 0
	printf("Top of Stack <0x%x> %p Current Stack <0x%x> %p Heap %p\n ",
		top_of_stack,
		top_of_stack,
		(uint32_t)&x, &x, p);
#endif
	if (p != NULL)
	{
		free(p);
	}

}
#endif

/* CCID support                                                                 */

#define CCID_MAX_APDUSIZE 255
#define STATUS_OFFSET 7
#define ERROR_OFFSET 8
#define dw2i(a, x) (unsigned int)((((((a[x+3] << 8) + a[x+2]) << 8) + a[x+1]) << 8) + a[x])

/* See CCID specs ch. 6.2.6, table 6.2.3 */
#define CCID_ICC_PRESENT_ACTIVE		0x00	/* 00 0000 00 */
#define CCID_ICC_PRESENT_INACTIVE	0x01	/* 00 0000 01 */
#define CCID_ICC_ABSENT				0x02	/* 00 0000 10 */
#define CCID_ICC_STATUS_MASK		0x03	/* 00 0000 11 */

#define CCID_COMMAND_FAILED			0x40	/* 01 0000 00 */
#define CCID_TIME_EXTENSION			0x80	/* 10 0000 00 */


/* write timeout
 * we don't have to wait a long time since the card was doing nothing           */
#define USB_WRITE_TIMEOUT (5 * 1000)	/* 5 seconds timeout */

/* This is the USB device with CCID capabilities, it must be a USB mass storage */
/* as well                                                                      */
/* As long as it is NULL it is not initialized and CCID cannot be used          */
/* Use initCCID() to initailize it                                              */
struct usb_device *udevCCID = NULL;

/* These are the endpoint descriptors                                           */
struct usb_endpoint_descriptor *epDescIn  = NULL;
struct usb_endpoint_descriptor *epDescOut = NULL;

/* Counter for USB read/write, taken from Linux CCID *ccid_descriptor->pbSeq    */
/* Set to 0 at any new APDU, incremented at each read/write operation           */
static unsigned char sequenceCount;

typedef enum _TARGET_DEVICE_TYPE {
	DEVICE_ASSD = 1,
	DEVICE_CCID
} TARGET_DEVICE_TYPE;

/* Target of dp_cmd, either goes to ASSD or to CCID, set with setTarget         */
static TARGET_DEVICE_TYPE targetType = DEVICE_ASSD;

/* Protcocol V1 handling                                                        */

typedef struct _COMMAND_TAG {
	uint8_t tag[4];
} __attribute__((packed)) COMMAND_TAG;

typedef struct _LENGTH_TAG {
	uint8_t length[4];
} __attribute__((packed)) LENGTH_TAG;


typedef struct _COMMAND_HEADER {
	uint8_t			version;
	COMMAND_TAG		cmdTag;
	LENGTH_TAG      lengthTag;
} __attribute__((packed)) COMMAND_HEADER;

#define LENGTH_HASHEDLOGIN_PAYLOAD_V1  96
#define LENGTH_POLL_PAYLOAD_V1         32
#define COMMAND_TAG_HASHEDLOGIN_V1     1
#define COMMAND_TAG_POLL_V1            0

/* Version 1 response tags according to the spec */
#define RESPONSE_TAG_WAIT_V1            0
#define RESPONSE_TAG_HASHEDLOGIN_V1     1
#define RESPONSE_TAG_ERROR_V1           2
#define RESPONSE_TAG_INFO_V1            3

/* When we receive RESPONSE_TAG_WAIT_V1 we set statusServer = STATUS_WAIT */
/* Once we received a valid packet we set  statusServer = STATUS_DONE   */
#define STATUS_DONE 1
#define STATUS_WAIT 2
unsigned statusServer;

/* When get_time(0) exceeds this value, waiting time has exceeded */
unsigned long end;

/* We need a larger buffer to send COMMAND_TAG_HASHEDLOGIN_V1 */
static uchar networkBuffer[2 * PKTSIZE_ALIGN + PKTALIGN];
/* Remember the buffer pointer to reset it                    */
static uchar *oldBuffer = NULL;

static const COMMAND_HEADER headerPoll = { 1,
								{ {0, 0, 0, COMMAND_TAG_POLL_V1} },
								{ {0, 0, 0, LENGTH_POLL_PAYLOAD_V1} }  };

static const COMMAND_HEADER headerHashedLogin = { 1,
						{ { 0, 0, 0, COMMAND_TAG_HASHEDLOGIN_V1 } },
						{ { 0, 0, 0, LENGTH_HASHEDLOGIN_PAYLOAD_V1 } } };
/* End protcocol V1 handling                                                   */

/**
 * DP MMC device index. Default value pulled from config.
 */
static int mmc_device = CONFIG_CMD_DP_MMC_DEVICE;

static int	usb_device			= 0;
static bool useUSBDevice		= false;
static bool validAnswerReceived = false;
static bool challengeTimeout	= false;
static bool abort				= false;

static char			serverAddr[256];
static int			serverPort;

#define VENDOR_MAGIC	0x63825363	/* RFC1048 Magic Cookie  */
#define PORT_BOOTPS	67		        /* BOOTP server UDP port */
#define PORT_BOOTPC	68		        /* BOOTP client UDP port */
#define TIMEOUT_MS	((3 + (5 * 5)) * 1000)

/* These are only relevant when we have a USB policy                            */
/* USB policy can have this format USB#1234#abcd                                */
static int vendorIdFromPolicy = 0;
static int productIdFromPolicy = 0;


static void dp_dhcp_request(void);

/* Current timeout handler */
static thand_f *time_handler;
/* Time base value */
static ulong	time_start;
/* Current timeout value */
static ulong	time_delta;

/* Counter of DHCP requests     */
static int			dhcp_try;

/* Start value of DHCP timer    */
static ulong		dhcp_start;

/* Start value of DP UDP timer  */
static ulong		dp_start;

/* DHCP timeout                 */
static ulong		dhcp_timeout;

/* DP UDP timeout               */
static ulong        dp_timeout;

/* Timeout limit                */
static ulong		time_taken_max = TIMEOUT_MS;

/* Transaction ids for DHCP		*/	
u32				dhcp_ids[4];
unsigned int	dhcp_num_ids;

/* DNS                          */
/* If the value in the configuration data is not an Ip address, we set it    */
/* to true */
static bool needDNSResolution = false;

/* Port we are expecting a DNS answer                                        */
static int dp_dns_our_port;

/* Last queried ip_addr                                                      */
static struct in_addr dp_ip_addr;
int transmitAPDU(uint8_t *APDU);

#ifdef PRINT_ASSD
static void mmmc_trace_ASSD(struct mmc *mmc, struct mmc_cmd *cmd,
					 int ret)
{
int i;
	u8 *ptr;

	if (ret) {
		printf("\t\tRET\t\t\t %d\n", ret);
	} else {
		switch (cmd->resp_type) {
		case MMC_RSP_NONE:
			printf("\t\tMMC_RSP_NONE\n");
			break;
		case MMC_RSP_R1:
			printf("\t\tMMC_RSP_R1,5,6,7 \t 0x%08X \n",
				cmd->response[0]);
			break;
		case MMC_RSP_R1b:
			printf("\t\tMMC_RSP_R1b\t\t 0x%08X \n",
				cmd->response[0]);
			break;
		case MMC_RSP_R2:
			printf("\t\tMMC_RSP_R2\t\t 0x%08X \n",
				cmd->response[0]);
			printf("\t\t          \t\t 0x%08X \n",
				cmd->response[1]);
			printf("\t\t          \t\t 0x%08X \n",
				cmd->response[2]);
			printf("\t\t          \t\t 0x%08X \n",
				cmd->response[3]);
			printf("\n");
			printf("\t\t\t\t\tDUMPING DATA\n");
			for (i = 0; i < 4; i++) {
				int j;
				printf("\t\t\t\t\t%03d - ", i*4);
				ptr = (u8 *)&cmd->response[i];
				ptr += 3;
				for (j = 0; j < 4; j++)
					printf("%02X ", *ptr--);
				printf("\n");
			}
			break;
		case MMC_RSP_R3:
			printf("\t\tMMC_RSP_R3,4\t\t 0x%08X \n",
				cmd->response[0]);
			break;
		default:
			printf("\t\tERROR MMC rsp not supported\n");
			break;
		}
	}
}
#endif

static void setTarget( const TARGET_DEVICE_TYPE type)
{
	targetType = type;
}

#ifdef DO_CONFIG
static const uint8_t MAGIC_CONF[] =
		{ 0x1E, 0x11, 0xA1, 0xBA, 0x87, 0x4A, 0xC6, 0xDB };
static const uint8_t VERSION[] = { 0x01 };
#endif

static const int NO_POLICY		= 0;
static const int PIN_POLICY		= 1;
static const int USB_POLICY		= 2;
static const int NET_POLICY		= 3;
static const int MAXPOLICY		= 4;

/**
 * Helper: get MMC device by index.
 */
static struct mmc *init_mmc_device(int dev, bool force_init)
{
	struct mmc *mmc;
	mmc = find_mmc_device(dev);
	if (!mmc) {
		printf("no mmc device at slot %x\n", dev);
		return NULL;
	}

	if (force_init)
		mmc->has_init = 0;
	if (mmc_init(mmc))
		return NULL;
	return mmc;
}

static void usbReset(void)
{
	/* When we reset USB, we must invalidate CCID as well                       */
	udevCCID = NULL;
	
	usb_stop();
	if (usb_init() < 0)
	{
		printf("usb_init failed");
		return;
	}
	usb_stor_scan(1);
}

/**
 * Force a device rescan. Needed for operations that cause a card reboot.
 */
static int dp_rescan(void)
{
	struct mmc *mmc;

	mmc = init_mmc_device(mmc_device, true);
	if (!mmc)
		return -1;

	return 0;
}

/*****************************************************************************
*
*					ReadUSB
*
****************************************************************************/
int ReadUSB(unsigned int *length, unsigned char *buffer)
{
	int rv;
	int actual_length;

	unsigned int pipe = usb_rcvbulkpipe(udevCCID, epDescIn->bEndpointAddress);

	rv = usb_bulk_msg(udevCCID,
		pipe,     // pipe
		buffer,									  // data	 
		*length,									  // len
		&actual_length,							  // actual_length
		USB_WRITE_TIMEOUT);						  // TIMEOUT	

	if (rv < 0)
	{

		return -1;
	}
	*length = actual_length;

	return 0;
} /* ReadUSB */



/*****************************************************************************
*
*					WriteUSB
*
****************************************************************************/
int WriteUSB( unsigned int length, unsigned char *buffer)
{
	int rv;
	int actual_length;

	unsigned int pipe = usb_sndbulkpipe(udevCCID, epDescOut->bEndpointAddress);

	rv = usb_bulk_msg(udevCCID,
		pipe,     
		buffer,									  // data	 
		length,									  // len
		&actual_length,							  // actual_length
		USB_WRITE_TIMEOUT);						  // TIMEOUT	

	if (rv < 0)
	{

		return -1;
	}

	return 0;
} /* WriteUSB */

  /*****************************************************************************
  *
  *					i2dw
  *
  ****************************************************************************/
static void i2dw(int value, unsigned char buffer[])
{
	buffer[0] = value & 0xFF;
	buffer[1] = (value >> 8) & 0xFF;
	buffer[2] = (value >> 16) & 0xFF;
	buffer[3] = (value >> 24) & 0xFF;
} /* i2dw */

  /*****************************************************************************
  *
  *					CCID_Receive
  *
  ****************************************************************************/
int CCID_Receive(unsigned int *rx_length,
	unsigned char rx_buffer[])
{
	unsigned char cmd[10 + CCID_MAX_APDUSIZE];			/* CCID + APDU buffer */
	unsigned int length;
	int return_value = 0 ;
	int ret;

time_request:
	length = sizeof(cmd);
	ret = ReadUSB( &length, cmd);
	if (0 != ret)
	{
		return ret;
	}

	if (length < STATUS_OFFSET + 1)
	{
		printf("Not enough data received: %d bytes", length);
		return -1;
	}

	if (cmd[STATUS_OFFSET] & CCID_COMMAND_FAILED)
	{
		switch (cmd[ERROR_OFFSET])
		{
		case 0xEF:	/* cancel */
			if (*rx_length < 2)
				return -1;
			rx_buffer[0] = 0x64;
			rx_buffer[1] = 0x01;
			*rx_length = 2;
			return 0;

		case 0xF0:	/* timeout */
			if (*rx_length < 2)
				return -1;
			rx_buffer[0] = 0x64;
			rx_buffer[1] = 0x00;
			*rx_length = 2;
			return 0;

		case 0xFD:	/* Parity error during exchange */
			return -1;

		default:
			return -1;
		}
	}

	if (cmd[STATUS_OFFSET] & CCID_TIME_EXTENSION)
	{
		printf("\nGot timeout\n");
		goto time_request;
	}

	/* we have read less (or more) data than the CCID frame says to contain */
	if (length - 10 != dw2i(cmd, 1))
	{
		printf("Can't read all data (%d out of %d expected)",
			length - 10, dw2i(cmd, 1));
		return_value = -1;
	}

	length = dw2i(cmd, 1);
	if (length <= *rx_length)
	{
		*rx_length = length;
	}
	else
	{
		printf("overrun by %d bytes", length - *rx_length);
		length = *rx_length;
		return_value = -1;
	}
	memcpy(rx_buffer, cmd + 10, length);
	return return_value;
} /* CCID_Receive */


/*****************************************************************************
 *
 *	CCID_Transmit, taken from Linux CCID
 *  we assume chain paarameter is always zero: all fits into one block
 *
****************************************************************************/

int CCID_Transmit(unsigned int tx_length,
	const unsigned char tx_buffer[])
{
	unsigned char cmd[10 + tx_length];	/* CCID + APDU buffer */
	int ret;

	/* See
	https://www.usb.org/sites/default/files/DWG_Smart-Card_CCID_Rev110.pdf
	p. 30 for the values below
	*/

	cmd[0] = 0x6F; /* XfrBlock */
	/* offset 1, 2, 3 and 4 are the APDU length                                 */
	i2dw(tx_length, cmd + 1);	/* APDU length */
	cmd[5] = 0;	/* slot number. always 0, since we support only one slot */
	cmd[6] = sequenceCount++;
	cmd[7] = 0;	/* No extension of block waiting timeout */

	cmd[8] = 0;
	cmd[9] = 0;

	memcpy(cmd + 10, tx_buffer, tx_length);

	ret = WriteUSB(10 + tx_length, cmd);

	return ret;
} /* CCID_Transmit */


/* Retrieves the device tree recursively                                        */
static bool findCCIDDevice(struct usb_device *dev)
{
	bool isCheckedDevice = true;

	if ((vendorIdFromPolicy != 0) && (productIdFromPolicy != 0)) 
	{
		isCheckedDevice = false;
		if ((dev->descriptor.idVendor == vendorIdFromPolicy) &&
			(dev->descriptor.idProduct == productIdFromPolicy))
		{
			isCheckedDevice = true;
		}
	}

	if (isCheckedDevice)
	{
		struct usb_config *config;
		struct usb_interface *ifdesc;
		struct usb_endpoint_descriptor *epdesc;
		int i, ii;

		config = &dev->config;
		for (i = 0; i < config->no_of_if; i++)
		{
			ifdesc = &config->if_desc[i];

			// Class has to be 0xb (CCID) and we must have two endpoints
			if ((0xb == ifdesc->desc.bInterfaceClass) &&
				(2 == ifdesc->no_of_ep)
				)
			{
				for (ii = 0; ii < ifdesc->no_of_ep; ii++)
				{
					epdesc = &ifdesc->ep_desc[ii];

					// Find out in and out endpoint
					if (epdesc->bEndpointAddress & 0x80)
					{
						epDescIn = epdesc;
					}
					else
					{
						epDescOut = epdesc;
					}
				}

				// Check whether found endpoints are really bulk endpoints
				if (2 == (epDescIn->bmAttributes & 0x03) &&
					2 == (epDescOut->bmAttributes & 0x03)
					)
				{
					udevCCID = dev;
					return true;
				}
			}
		}
	}

#ifdef CONFIG_DM_USB 
	struct udevice *child;

	for (device_find_first_child(dev->dev, &child);
		child;
		device_find_next_child(&child)) {
		struct usb_device *udev;

		if (!device_active(child))
			continue;

		udev = dev_get_parent_priv(child);

		/*
		* we only want mass storage devices, ignore anything else
		*/
		if ((device_get_uclass_id(child) == UCLASS_MASS_STORAGE) ||
			(device_get_uclass_id(child) == UCLASS_USB_HUB))
		{
			if (findCCIDDevice(udev))
			{
				return true;
			}
		}
	}
#else
	if (dev->maxchild > 0) {
		for (int i = 0; i < dev->maxchild; i++) {
			if (dev->children[i] != NULL)
			{
				if (findCCIDDevice(dev->children[i]))
				{
					return true;
				}
			}
		}
	}
#endif

	return false;
}


/**
  * Try to find and initialize CCID
  * sets udevCCID, epDescIn, epDescOut 
  *  
  *
  * @returns true on success, false on failure.
*/
static bool initCCID(void)
{
	bool hasCCIDDevice = false;

#ifdef CONFIG_DM_USB
	struct udevice *bus;

	for (uclass_find_first_device(UCLASS_USB, &bus);
		bus;
		uclass_find_next_device(&bus))
	{
		struct usb_device *udev;
		struct udevice *dev;

		if (!device_active(bus))
			continue;

		device_find_first_child(bus, &dev);
		if (dev && device_active(dev))
		{
			udev = dev_get_parent_priv(dev);
			hasCCIDDevice = findCCIDDevice(udev);
		}
		if (hasCCIDDevice)
		{
			break;
		}
	}
#else
	struct usb_device *udev;
	int i;
	usbReset();
	for (i = 0; i < USB_MAX_DEVICE; i++) {
		udev = usb_get_dev_index(i);
		if (udev == NULL)
			break;
		if (udev->parent == NULL)
			hasCCIDDevice = findCCIDDevice(udev);
		if (hasCCIDDevice)
		{
			break;
		}
	}
#endif

	return hasCCIDDevice;
}

/* cmdCCID()                                                                    */
/* Send APDU through to the CCID device and retrieve response                   */
/* This is taken from CmdXfrBlockAPDU_extended in Linux CCID                    */
/* Support for APDU chaining removed                                            */
/* @returns 0 on success, -1 on failure                                         */
static int cmdCCID(
	unsigned int tx_length, 
	const unsigned char tx_buffer[], unsigned int *rx_length,
	unsigned char rx_buffer[])
{
	int return_value;
	unsigned int local_rx_length = *rx_length;;

	/* TODO: find the max CCID message length, it is somewhere in the CCID      */
	return_value = CCID_Transmit(tx_length, tx_buffer);

	if (return_value != 0)
	{
		return return_value;
	}

	/* We assume response fits into one block, ie. chain_parameter  is
	always 0, and no buffer overflow
	*/
	return_value = CCID_Receive(&local_rx_length, rx_buffer);

	/* Error occured, bail out                                                  */
	if (return_value != 0)
	{
		return return_value;
	}

	*rx_length = local_rx_length;

	/* Success                                                                  */
	return 0;
} 

/**
* Send C-APDU to CCID device and return R-APDU.
*
* @returns Length of R-APDU on success, -1 on failure.
*/
static ssize_t dp_cmdCCID(const uint8_t *request, size_t request_len,
	uint8_t *response, size_t response_len)
{
	int ret = 0;

#ifdef PRINT_VERBOSE
	printf("C-APDU CCID:\n");
	print_buffer(0x0000000, request, 1, request_len, 0);
#endif
	unsigned int rx_length = response_len;
	ret = cmdCCID( request_len, request, &rx_length, response);
	if (ret != 0)
	{
		/* Error occured                                                        */
		return -1;
	}
#ifdef PRINT_VERBOSE
	printf("R-APDU CCID:\n");
	print_buffer(0x0000000, response, 1, rx_length, 0);
#endif
	return rx_length;
}

/**
* Send C-APDU through ASSD and return R-APDU.
*
* @returns Length of R-APDU on success, -1 on failure.
*/
static ssize_t dp_cmdASSD(const uint8_t *request, size_t request_len,
		      uint8_t *response, size_t response_len)
{
	uint8_t block[512] __attribute__ ((aligned(CONFIG_SYS_CACHELINE_SIZE)));
	uint16_t length;
	struct mmc *mmc;
	int err;

	mmc = init_mmc_device(mmc_device, false);


	if (!mmc)
		return -1;

#ifdef PRINT_VERBOSE
	printf("C-APDU:\n");
	print_buffer(0x0000000, request, 1, request_len, 0);
#endif

	/* Build and send request block. */
	request_len = request_len ;
	size_t request_lenplus = request_len + 2;

	length = htons((uint16_t)request_lenplus);
	
	// Create ASSD Command	
	memset(block, 0x00, sizeof(block));	
	memcpy(block, &length, sizeof(length)); // 2 byte length of command
	memcpy(block+sizeof(length), request, request_len); // APDU 

#ifdef PRINT_ASSD
	print_buffer(0x0000000, block, 1, 512, 0);
#endif
	err = transmitAPDU(block);

	if (err !=0)
	{
		return err;
	}
		
	/* Read length. */
	memcpy(&length, block ,sizeof(length));
	length = ntohs(length);
	length = length - 2; // ASSD length includes length bytes in length

	/* Make sure there's enough response buffer. */
	if (length > response_len) {
		printf("Response too large to fit in buffer!\n");
		return -1;
	}

	/* Copy and return response. */
	memcpy(response,block+2, length);

#ifdef PRINT_VERBOSE
	printf("R-APDU:\n");
	print_buffer(0x0000000, response, 1, length, 0);
#endif

	return length;
}

/**
* Send C-APDU and return R-APDU.
*
* @returns Length of R-APDU on success, -1 on failure.
*/
static ssize_t dp_cmd(const uint8_t *request, size_t request_len,
	uint8_t *response, size_t response_len)
{
	switch (targetType)
	{
	case DEVICE_ASSD:
		return dp_cmdASSD(request, request_len, response, response_len);
		break;
	case DEVICE_CCID:
		if (NULL == udevCCID)
		{
			initCCID();
		}
		if (NULL != udevCCID)
		{
			return dp_cmdCCID(request, request_len, response, response_len);
		}
		break;
	default:
		break;
	}
	return 0;
}


static ssize_t unpack_hex_string(uint8_t *bin, const char *hex, size_t maxlen)
{
	if (strlen(hex) % 2) {
		printf("hex string length not even!\n");
		return -1;
	}
	if (strlen(hex)/2 > maxlen) {
		printf("hex string too long to fit in buffer!\n");
		return -1;
	}

	ssize_t length = 0;

	while (*hex) {
		char b[3];
		b[0] = *hex++;
		b[1] = *hex++;
		b[2] = '\0';
		*bin++ = simple_strtoul(b, NULL, 16);
		length++;
	}

	return length;
}

static int do_dp_dev(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	if (argc == 1) {
		printf("Current MMC device: %d\n", mmc_device);
	} else if (argc == 2) {
		mmc_device = simple_strtoul(argv[1], NULL, 10);
		printf("MMC device is now: %d\n", mmc_device);
	}

	return CMD_RET_SUCCESS;
}

static int printUSBStatus(void)
{
	usb_stor_info();

	if (useUSBDevice)
	{
		printf("Current USB device (selected): %d\n", usb_device);
	}
	else
	{
		printf("Current USB device (not selected): %d\n", usb_device);
	}
	return CMD_RET_SUCCESS;
}

/*
 *		dp usb		: do a reset of the USB subsystem
 *		dp usb 1	: select USB as destination of write and read
 *		dp usb 0    : select MMC as destination of write and read
 *		dp usb 1 0  : dp usb 1 and set device 0
*/
static int do_dp_usb(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	if (argc == 1) 
	{
		usbReset();
	}
	else if (argc == 2)
	{
		int arg = simple_strtoul(argv[1], NULL, 10);
		if (1 == arg)
		{
			useUSBDevice = true;
		}
		else
		{
			useUSBDevice = false;
		}
	}
	else if (argc == 3) 
	{
		int arg = simple_strtoul(argv[1], NULL, 10);
		if (1 == arg)
		{
			useUSBDevice = true;
		}
		else
		{
			useUSBDevice = false;
		}
		usb_device = simple_strtoul(argv[2], NULL, 10);
	}

	printUSBStatus();
	return CMD_RET_SUCCESS;
}


static int do_dp_cmd(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	uint8_t request[256];
	size_t request_len;
	uint8_t response[256];
	ssize_t response_len;

	if (argc != 2)
		return CMD_RET_USAGE;

	if (strlen(argv[1]) % 2) {
		printf("APDU string length not even!\n");
		return CMD_RET_FAILURE;
	}

	/* Parse APDU. */
	request_len = unpack_hex_string(request, argv[1], sizeof(request));
	if (request_len == -1) {
		return CMD_RET_FAILURE;
	}

	/* Execute APDU. */
	response_len = dp_cmd(request, request_len, response, sizeof(response));
	if (response_len < 0) {
		return CMD_RET_FAILURE;
	}

	return CMD_RET_FAILURE;
}


static int dp_status( STATUS_RESPONSE *pres)
{
	static const uint8_t request[] = { 0xff, 0x70, 0x00, 0x00, 0x00 };

	ssize_t length = dp_cmd(request, sizeof(request), (void *)pres,
		sizeof(STATUS_RESPONSE));
	if (length != sizeof(STATUS_RESPONSE))
		return CMD_RET_FAILURE;

	if (ntohs( (pres->status) ) != 0x9000)
		return CMD_RET_FAILURE;

	return CMD_RET_SUCCESS;
}


static int do_dp_status(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	STATUS_RESPONSE response;
	int res = CMD_RET_SUCCESS;

	res = dp_status(&response);
	if (CMD_RET_SUCCESS != res)
	{
		return res;
	}

	printf("  Card Flash Encryption Mode: 0x%02x\n"
	       "  System state: 0x%02x\n"
	       "  Password retry count: %d\n"
	       "  SO Password retry count: %d\n"
	       "  Number of Resets: %u\n"
	       "  CD-ROM address: 0x%08x\n"
	       "  Extended Security flags: 0x%02x\n",
	       response.mode, response.state,
	       response.counter, response.so_counter,
	       ntohl(response.reset_counter),
	       ntohl(response.cdrom),
	       response.flags);

	/* Copy status block into memory if requested. */
	if (argc >= 2) {
		void *dest = (void *)simple_strtoul(argv[1], NULL, 16);
		memcpy(dest, &response, sizeof(response)-2);
	}

	printf("OK.\n");
	return CMD_RET_SUCCESS; 
}

static int dp_nvramstatus(STATUS_NVRAMRESPONSE *pres)
{
	static const uint8_t request[] = { 0xff, 0x70, 0x03, 0x00, 0x00 };

	ssize_t length = dp_cmd(request, sizeof(request), (void *)pres,
		sizeof(STATUS_NVRAMRESPONSE));
	if (length != sizeof(STATUS_NVRAMRESPONSE))
		return CMD_RET_FAILURE;

	if (ntohs((pres->status)) != 0x9000)
		return CMD_RET_FAILURE;

	return CMD_RET_SUCCESS;
}


static int do_dp_nvramstatus(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	STATUS_NVRAMRESPONSE response;
	int res = CMD_RET_SUCCESS;

	res = dp_nvramstatus(&response);
	if (CMD_RET_SUCCESS != res)
	{
		return res;
	}

	printf("  Hiden sectors: 0x%08x\n"
		"  Random rights: 0x%02x\n"
		"  Cylic rights: 0x%02x\n"
		"  Random sectors: 0x%08x\n"
		"  Cyclic sectors: 0x%08x\n"
		"  Next sector: 0x%08x\n"
		"  Wrap: 0x%02x\n",
		 ntohl(response.hiddenSectors),
		 response.accessRandomRights,
		 response.accessCyclicRights,
		 ntohl(response.randomSectors),
		 ntohl(response.cyclicSectors),
		 ntohl(response.nextSector),
		 response.wrap);

	/* Copy status block into memory if requested. */
	if (argc >= 2) {
		void *dest = (void *)simple_strtoul(argv[1], NULL, 16);
		memcpy(dest, &response, sizeof(response) - 2);
	}

	printf("OK.\n");
	return CMD_RET_SUCCESS;
}

static int dp_configureNVRAM(const uint8_t randomRights,
	const  uint8_t  cyclicRights,
	const  uint32_t  numRandomSectors,
	const  uint32_t  numCyclicSectors,
	const bool doFuse)
{
	uint8_t request[15];
	uint8_t *rq = request;

	*rq++ = 0xff;
	*rq++ = 0x80;
	*rq++ = 0x03;
	if (doFuse)
	{
		*rq++ = 0x1;
	}
	else
	{
		*rq++ = 0x0;
	}
	*rq++ = 0x0a;
	*rq++ = randomRights;
	*rq++ = cyclicRights;

	*rq++ = (numRandomSectors >> 24) & 0xFF;
	*rq++ = (numRandomSectors >> 16) & 0xFF;
	*rq++ = (numRandomSectors >> 8) & 0xFF;
	*rq++ = (numRandomSectors) & 0xFF;

	*rq++ = (numCyclicSectors >> 24) & 0xFF;
	*rq++ = (numCyclicSectors >> 16) & 0xFF;
	*rq++ = (numCyclicSectors >> 8) & 0xFF;
	*rq++ = (numCyclicSectors) & 0xFF;

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *)&request, request[4] + 5, response,
		sizeof(response));
	if (length != 2) {
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00)) {
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

/* Get size of random NVRAM block 0 */
static int dp_readConfigSize(uint16_t *size)
{
	uint8_t request[9];
	uint8_t *rq = request;

	*rq++ = 0xff;
	*rq++ = 0xd0;
	// Use random mem
	*rq++ = 0x00;
	// Return size
	*rq++ = 0x01;
	*rq++ = 0x04;
	// Offset zero
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;

	/* Execute request. */
	struct {
		uint8_t size[2];
		uint16_t code;
	} __attribute__((packed)) response;

	ssize_t length = dp_cmd((void *)&request, request[4] + 5,
		(void *)&response,
		sizeof(response));
	if (length != 4)
	{
		return CMD_RET_FAILURE;
	}
	if (ntohs(response.code) != 0x9000) {
		printf("unexpected return code\n");
		return CMD_RET_FAILURE;
	}

	*size = ((response.size[0] & 0xff) << 8) | (response.size[1] & 0xff);

	return CMD_RET_SUCCESS;
}

static int dp_readConfig(uint8_t *data, const uint16_t expectedSize)
{
	uint8_t request[9];
	uint8_t *rq = request;

	*rq++ = 0xff;
	*rq++ = 0xd0;
	// Use random mem
	*rq++ = 0x00;
	// Return data
	*rq++ = 0x00;
	*rq++ = 0x04;
	// Offset zero
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;

	/* Execute request. */
	uint8_t response[expectedSize + 2];
	ssize_t length = dp_cmd((void *)&request, request[4] + 5, response,
		sizeof(response));
	if (length != (expectedSize + 2))
	{
		printf("dp_readConfig failed; Size %d %d\n", (int)length, 
										(int)expectedSize);
		return CMD_RET_FAILURE;
	}
	if (!(response[expectedSize] == 0x90 && 
		response[expectedSize +1 ] == 0x00)) 
	{
		printf("dp_readConfig failed: result\n");
		return CMD_RET_FAILURE;
	}

	memcpy(data, response, expectedSize);

	return CMD_RET_SUCCESS;
}

static int do_dp_readConfig(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	int ret = CMD_RET_SUCCESS;

	if (argc != 2)
	{
		return CMD_RET_USAGE;
	}

	uint16_t dataSize = (uint16_t)simple_strtoul(argv[1], NULL, 16);
	uint8_t data[dataSize];

	ret = dp_readConfig(data, dataSize);

	if (CMD_RET_SUCCESS != ret)
	{
		printf("dp_readConfig failed\n");
		return CMD_RET_FAILURE;
	}

#ifdef PRINT_VERBOSE
	print_buffer(0x0000000, data, 1, dataSize, 0);
	printf("OK.\n");
#endif

	return CMD_RET_SUCCESS;

}


static int dp_writeConfig(const uint8_t *data, const uint16_t dataSize, 
										bool doAppend)
{
	uint8_t request[dataSize + 4 + 5];
	uint8_t *rq = request;

	*rq++ = 0xff;
	// Write
	*rq++ = 0xd1;
	// Use random mem
	*rq++ = 0x00;
	if (doAppend)
	{
		*rq++ = 0x01;
	}
	else
	{
		*rq++ = 0x00;
	}

	*rq++ = 0x04 + dataSize;
	// Offset zero
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;

	memcpy(rq, data, dataSize);

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *)&request, request[4] + 5, response,
		sizeof(response));
	if (length != 2) {
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 &&
		response[1] == 0x00))
	{
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

#ifdef DO_CONFIG
static int write_defaultPolicy(void)
{
	int res = CMD_RET_SUCCESS;

	res = dp_writeConfig(MAGIC_CONF, sizeof(MAGIC_CONF), false);
	if (CMD_RET_SUCCESS != res)
	{
		goto done;
	}

	res = dp_writeConfig(VERSION, sizeof(VERSION), true);
	if (CMD_RET_SUCCESS != res)
	{
		goto done;
	}
	
	res = dp_writeConfig(DEFAULT_POLICY, sizeof(DEFAULT_POLICY), true);
	if (CMD_RET_SUCCESS != res)
	{
		goto done;
	}

done:
	return res;
}
#endif

static int do_dp_writeConfig(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	int res = CMD_RET_SUCCESS;

	if (argc != 2)
	{
		return CMD_RET_USAGE;
	}

	res = dp_writeConfig((uint8_t*)argv[1], (uint16_t)strlen(argv[1]), false);

	if (CMD_RET_SUCCESS != res)
	{
		return CMD_RET_FAILURE;
	}

	printf("OK.\n");

	return CMD_RET_SUCCESS;
}

static int dp_sizeConfig(uint16_t *size)
{
	uint8_t request[9];
	uint8_t *rq = request;

	*rq++ = 0xff;
	*rq++ = 0xd0;
	// Use random mem
	*rq++ = 0x00;
	// Return size
	*rq++ = 0x1;
	*rq++ = 0x04;
	// Offset zero
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x00;

	/* Execute request. */
	uint8_t response[4];
	ssize_t length = dp_cmd((void *)&request, request[4] + 5, response,
		sizeof(response));
	if (length != 4) {
		return CMD_RET_FAILURE;
	}
	if (!(response[2] == 0x90 && response[3] == 0x00)) {
		return CMD_RET_FAILURE;
	}

	*size = ((response[0] & 0xFF) << 8) | (response[1] & 0xFF);


	return CMD_RET_SUCCESS;
}

static int do_dp_sizeConfig(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	uint16_t dataSize = 0;

	int res = CMD_RET_SUCCESS;

	if (argc != 1)
	{
		return CMD_RET_USAGE;
	}

	res = dp_sizeConfig(&dataSize);

	if (CMD_RET_SUCCESS != res)
	{
		return CMD_RET_FAILURE;
	}

#ifdef PRINT_VERBOSE
	printf("Config area has 0x%08x byte of data\n", dataSize);
	printf("OK.\n");
#endif

	return CMD_RET_SUCCESS;
}

static int do_dp_configureNVRAM(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	uint8_t		randomRights;
	uint8_t		cyclicRights;
	uint32_t	numRandomSectors;
	uint32_t	numCyclicSectors;
	int			res = CMD_RET_SUCCESS;

	if (argc != 5)
	{
		return CMD_RET_USAGE;
	}
	randomRights	 = (uint8_t)((simple_strtoul(argv[1], NULL, 16)) & 0xff);
	cyclicRights	 = (uint8_t)((simple_strtoul(argv[2], NULL, 16)) & 0xff);
	numRandomSectors = simple_strtoul(argv[3], NULL, 16);
	numCyclicSectors = simple_strtoul(argv[4], NULL, 16);

	// Never fuse from the command line
	res = dp_configureNVRAM(randomRights, cyclicRights, 
							numRandomSectors, numCyclicSectors, false);

	if (CMD_RET_SUCCESS != res)
	{
		return CMD_RET_FAILURE;
	}

	printf("do_dp_configureNVRAM succeeded.\n");
	return CMD_RET_SUCCESS;
}


static int dp_activate(uint8_t *pwd, ssize_t pwdLen, uint8_t *soPwd,
	ssize_t soLen, int retryCounter)
{
	if ( (NULL == pwd) || (NULL == soPwd) )
	{
		return CMD_RET_FAILURE;
	}
	uint8_t request[72];
	uint8_t *rq = request;

	*rq++ = 0xff;
	*rq++ = 0x10;
	*rq++ = 0x0;
	*rq++ = 0x0;
	*rq++ = pwdLen + soLen + 3;

	*rq++ = (uint8_t)pwdLen;
	memcpy(rq, pwd, pwdLen);
	rq += pwdLen;

	*rq++ = (uint8_t)soLen;
	memcpy(rq, soPwd, soLen);
	rq += soLen;

	*rq = (uint8_t)retryCounter;

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *)&request, request[4] + 5, response,
		sizeof(response));
	if (length != 2) {
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00)) {
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}


static int do_dp_activate(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	ssize_t len				= 0;
	ssize_t pwdLen			= 0;
	ssize_t soLen			= 0;
	int		retryCounter	= 0;
	uint8_t pwd[DP_PASSWORD_MAX_LENGTH];
	uint8_t soPwd[DP_PASSWORD_MAX_LENGTH];
	int res = CMD_RET_SUCCESS;

	if (argc != 4)
	{
		return CMD_RET_USAGE;
	}

	len = strlen(argv[1]);
	if (len % 1)
	{
		return CMD_RET_FAILURE;
	}

	pwdLen = len / 2;
	if (pwdLen > DP_PASSWORD_MAX_LENGTH)
	{
		return CMD_RET_FAILURE;
	}

	/* Unpack arguments. */
	unpack_hex_string(pwd, argv[1], DP_PASSWORD_MAX_LENGTH);

	len = strlen(argv[2]);
	if (len % 1)
	{
		return CMD_RET_FAILURE;
	}

	soLen = len / 2;
	if (soLen > DP_PASSWORD_MAX_LENGTH)
	{
		return CMD_RET_FAILURE;
	}

	unpack_hex_string(soPwd, argv[2], DP_PASSWORD_MAX_LENGTH);
												
	retryCounter = simple_strtoul(argv[3], NULL, 10);
	if (retryCounter < 3 || retryCounter > 15) {
		printf("retry counter outside allowed range (3..15)\n");
		return CMD_RET_FAILURE;
	}

	res = dp_activate(pwd, pwdLen, soPwd, soLen, retryCounter);

	if (CMD_RET_SUCCESS != res)
	{
		return CMD_RET_FAILURE;
	}
	       
	printf("Activate succeded.\n");
	return CMD_RET_SUCCESS;
}

static int do_dp_deactivate(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	ssize_t len = 0;
	ssize_t pwdLen = 0;

	uint8_t request[38];
	uint8_t *rq = request;

	if (argc != 2)
	{
		return CMD_RET_USAGE;
	}

	// CMD 
	*rq++ = 0xff;
	*rq++ = 0x20;
	*rq++ = 0x0;
	*rq++ = 0x0;
	*rq++ = 0x21;

	len = strlen(argv[1]);
	if (len % 1)
	{
		return CMD_RET_FAILURE;
	}

	pwdLen = len / 2;
	if (pwdLen > DP_PASSWORD_MAX_LENGTH)
	{
		return CMD_RET_FAILURE;
	}
	*rq++ = (uint8_t)pwdLen;

	/* Unpack arguments. */
	unpack_hex_string(rq, argv[1], DP_PASSWORD_MAX_LENGTH);
	rq += pwdLen;

	request[4] = 1 + pwdLen;

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *) &request, request[4] + 5, response,
															sizeof(response));
	if (length != 2) {
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00)) {
		return CMD_RET_FAILURE;
	}
	       
	printf("Dectivate succeded.\n");
	return CMD_RET_SUCCESS;
}

static int getLoginChallenge(uint8_t *challenge)
{
	int ret = CMD_RET_SUCCESS;

	static const uint8_t challenge_request[5] =
	{ 0xff, 0x70, 0x05, 0x00, 0x00 };
	struct {
		uint8_t challenge[DP_CHALLENGE_LENGTH];
		uint16_t code;
	} __attribute__((packed)) challenge_response;

	if (dp_cmd((void *)&challenge_request, sizeof(challenge_request),
		(void *)&challenge_response, sizeof(challenge_response)) !=
		sizeof(challenge_response))
	{
		printf(
			"Failed to retrieve challenge from card in SD card slot: unexpected return length\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}

	if (ntohs(challenge_response.code) != 0x9000)
	{
		printf(
			"Failed to retrieve challenge from card in SD card slot: unexpected return code\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}

	memcpy(challenge, challenge_response.challenge, DP_CHALLENGE_LENGTH);

failed:
	return ret;

}


static int dp_hashedverify(uint8_t *pwd, ssize_t pwdLen)
{
	int ret;

	/* Retrieve challenge. */
	uint8_t challenge[DP_CHALLENGE_LENGTH];
	ret = getLoginChallenge(challenge);
	if (CMD_RET_SUCCESS != ret)
	{
		return ret;
	}

	/* Compute hash = sha256(sha256(pin)+challenge) */
	uint8_t hash[SHA256_SUM_LEN];
	sha256_context ctx;

	sha256_starts(&ctx);
	sha256_update(&ctx, pwd, pwdLen);
	sha256_finish(&ctx, hash);

	sha256_starts(&ctx);
	sha256_update(&ctx, hash, sizeof(hash));
	sha256_update(&ctx, challenge, sizeof(challenge));
	sha256_finish(&ctx, hash);

	/* Send hash. */
	uint8_t ins = 0x30;   // PIN login
	struct {
		uint8_t header[5];
		uint8_t length;
		uint8_t hash[DP_HASH_LENGTH];
	} __attribute__((packed)) hash_request = {
		.header = { 0xff, ins, 0x00, 0x00, 0x21 },
		.length = 0x20,
		.hash = { 0 },
	};
	uint8_t hash_response[2];

	memcpy(hash_request.hash, hash, sizeof(hash_request.hash));

	if (dp_cmd((void *)&hash_request, sizeof(hash_request),
		(void *)&hash_response, sizeof(hash_response)) !=
		sizeof(hash_response)) {
		printf("unexpected return length\n");
		return CMD_RET_FAILURE;
	}
	if (!(hash_response[0] == 0x90 && hash_response[1] == 0x00)) {
		printf("unexpected return code\n");
		return CMD_RET_FAILURE;
	}

	printf("OK.\n");
	return CMD_RET_SUCCESS;
}


static int dp_verify(uint8_t *pwd, ssize_t pwdLen)
{
	uint8_t request[38];
	uint8_t *rq = request;
	uint8_t response[2];

	// CMD 
	*rq++ = 0xff;
	*rq++ = 0x30;
	*rq++ = 0x0;
	*rq++ = 0x0;
	*rq++ = 1 + pwdLen;
	*rq++ = (uint8_t)pwdLen;

	memcpy(rq, pwd, pwdLen);

	if (dp_cmd((void *)&request, request[4] + 5,
		(void *)&response, sizeof(response)) !=
		sizeof(response)) {
		printf("unexpected return length\n");
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00)) {
		printf("unexpected return code\n");
		return CMD_RET_FAILURE;
	}
	return CMD_RET_SUCCESS;
}

static int do_dp_verify(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	uint8_t pwd[DP_PASSWORD_MAX_LENGTH];


	if (argc != 2)
		return CMD_RET_USAGE;

	int res = CMD_RET_SUCCESS;
	ssize_t len = 0;
	ssize_t pwdLen = 0;


	len = strlen(argv[1]);
	if (len % 1)
	{
		return CMD_RET_FAILURE;
	}

	pwdLen = len / 2;
	if (pwdLen > DP_PASSWORD_MAX_LENGTH)
	{
		return CMD_RET_FAILURE;
	}

	/* Unpack arguments. */
	unpack_hex_string(pwd, argv[1], DP_PASSWORD_MAX_LENGTH);

	res = dp_verify(pwd, pwdLen);

	if (CMD_RET_SUCCESS != res)
	{
		return CMD_RET_FAILURE;
	}

	printf("Verify OK.\n");
	return CMD_RET_SUCCESS;
}

static int do_dp_hashed(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	int ret;

	if (argc != 3)
		return CMD_RET_USAGE;

	uint8_t ins;
	if (!strcmp(argv[1], "PIN")) {
		ins = 0x30;
	} else if (!strcmp(argv[1], "SO-PIN")) {
		ins = 0x20;
	} else {
		return CMD_RET_USAGE;
	}

	uint8_t password[DP_PASSWORD_LENGTH];
//	if (unpack_hex_string(password, argv[2], 
//								sizeof(password)) != DP_PASSWORD_LENGTH) {
//		printf("only full-length passwords are supported\n");
//		return CMD_RET_FAILURE;
//	}
	ssize_t passwordLength = unpack_hex_string(password, argv[2],
													sizeof(password));


	uint8_t challenge[DP_CHALLENGE_LENGTH];
	ret = getLoginChallenge(challenge);
	if (CMD_RET_SUCCESS != ret)
	{
		return ret;
	}

	/* Compute hash = sha256(sha256(pin)+challenge) */
	uint8_t hash[SHA256_SUM_LEN];
	sha256_context ctx;

	sha256_starts(&ctx);
//	sha256_update(&ctx, password, sizeof(password));
	
	sha256_update(&ctx, password, passwordLength);
	sha256_finish(&ctx, hash);

	sha256_starts(&ctx);
	sha256_update(&ctx, hash, sizeof(hash));
	sha256_update(&ctx, challenge, sizeof(challenge));
	sha256_finish(&ctx, hash);

	/* Send hash. */
	struct {
		uint8_t header[5];
		uint8_t length;
		uint8_t hash[DP_HASH_LENGTH];
	} __attribute__ ((packed)) hash_request = {
		.header = { 0xff, ins, 0x00, 0x00, 0x21 },
		.length = 0x20,
		.hash = { 0 },
	};
	uint8_t hash_response[2];

	memcpy(hash_request.hash, hash, sizeof(hash_request.hash));

	if (dp_cmd((void *) &hash_request, sizeof(hash_request),
	           (void *) &hash_response, sizeof(hash_response)) !=
	          sizeof(hash_response)) {
		printf("unexpected return length\n");
		return CMD_RET_FAILURE;
	}
	if (!(hash_response[0] == 0x90 && hash_response[1] == 0x00)) {
		printf("unexpected return code\n");
		return CMD_RET_FAILURE;
	}

	printf("OK.\n");
	return CMD_RET_SUCCESS;
}

static int do_dp_extended(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	if (argc != 2)
		return CMD_RET_USAGE;

	uint8_t request[] = { 0xff, 0x80, 0x05, 0x00, 0x01, 0x00 };
	request[5] = simple_strtoul(argv[1], NULL, 16);

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *) &request, sizeof(request), response, 
															sizeof(response));

	/* Kludge: card reboots on success. Rescan MMC bus and pretend
	 * we've succeeded if this happens. */
	if (length == -1) {
		printf("Card rebooted; reinitializing MMC.\n");
		if (dp_rescan() == 0) {
			printf("Card rebooted successfully.\n");
			return CMD_RET_SUCCESS;
		} else {
			printf("Card failed to come back after reboot. Is it toast?\n");
			return CMD_RET_FAILURE;
		}
	}

	if (length != 2) {
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00)) {
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

static int do_dp_id(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	uint8_t request[] = { 0xff, 0x70, 0x01, 0x00, 0x00 };

	/* Execute request. */
	uint8_t response[DP_ID_LENGTH+2];
	ssize_t length = dp_cmd((void *) &request, sizeof(request), response, 
															sizeof(response));
	if (length != sizeof(response)) {
		return CMD_RET_FAILURE;
	}
	if (!(response[DP_ID_LENGTH] == 0x90 && response[DP_ID_LENGTH+1] == 0x00))
	{
		return CMD_RET_FAILURE;
	}

	/* Output ID to console. */
	printf("Card ID:\n");
	print_buffer(0x0000000, response, 1, DP_ID_LENGTH, 0);

	/* Save ID to memory if requested. */
	if (argc >= 2)
	{
		void *dest = (void *)simple_strtoul(argv[1], NULL, 16);
		memcpy(dest, response, DP_ID_LENGTH);
	}

	return CMD_RET_SUCCESS;
}

static int do_dp_cdrom(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	if (argc != 2) {
		return CMD_RET_USAGE;
	}

	uint8_t request[5] = { 0xff, 0x53, 0x00, 0x00, 0x00 };
	request[2] = simple_strtoul(argv[1], NULL, 10);

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *) &request, sizeof(request), response, 
															sizeof(response));
	if (length != sizeof(response)) 
	{
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00))
	{
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

static int do_dp_exception(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	if (argc != 3)
	{
		return CMD_RET_USAGE;
	}

	struct {
		uint8_t header[5];
		uint32_t write;
		uint32_t read;
	} __attribute__ ((packed)) request = {
		.header = { 0xff, 0x53, 0x02, 0x00, 0x08 },
	};
	request.write = htonl(simple_strtoul(argv[1], NULL, 16));
	request.read  = htonl(simple_strtoul(argv[2], NULL, 16));

	/* Execute request. */
	uint8_t response[2];
	ssize_t length = dp_cmd((void *) &request, sizeof(request), response, 
															sizeof(response));
	if (length != sizeof(response)) 
	{
		return CMD_RET_FAILURE;
	}
	if (!(response[0] == 0x90 && response[1] == 0x00))
	{
		return CMD_RET_FAILURE;
	}

	return CMD_RET_SUCCESS;
}

#ifdef DO_CONFIG
static bool checkMagicData(const uint8_t *policyData)
{
	bool hasMagicData = false;
	for (int j = 0; j < sizeof(MAGIC_CONF); j++)
	{
		if (*(policyData + j) != *(MAGIC_CONF + j))
		{
			goto failed;
		}
	}
	hasMagicData = true;
failed:
	return hasMagicData;
}
#endif

static int doHashedPIN(void)
{
	// Assumes card state is 2: activated and locked

	int ret = CMD_RET_SUCCESS;

	int res = 0;
	int count = 0;
	char pwd[DP_PASSWORD_MAX_LENGTH + 1];

	// Ask for PIN
	printf("\n Enter your PIN: ");

	memset(pwd, 0, sizeof(pwd));
	do
	{
		res = getc();
		if (res != 13)
		{
			pwd[count] = (char)res;
			printf("*");
		}
		count++;
	} while (res != 13 && res >= 0 && count < DP_PASSWORD_MAX_LENGTH);

	printf("\n");

#ifdef PRINT_VERBOSE
	printf("You entered <%s>\n", pwd);
#endif

	ret = dp_hashedverify((uint8_t *)pwd, strlen(pwd));
	memset(pwd, 0, sizeof(pwd));

	return ret;
}


static int doPIN(void)
{
	/* Assumes card state is 2: activated and locked */
	int ret = CMD_RET_SUCCESS;

	int res = 0;
	int count = 0;
	char pwd[DP_PASSWORD_MAX_LENGTH + 1];

	// Ask for PIN
	printf("\n Enter your PIN: ");

	memset(pwd, 0, sizeof(pwd));
	do
	{
		res = getc();
		if (res != 13)
		{
			pwd[count] = (char)res;
			printf("*");
		}
		count++;
	} while (res != 13 && res >= 0 && count < DP_PASSWORD_MAX_LENGTH);
	
	printf("\n");

#ifdef PRINT_VERBOSE
	printf("You entered <%s>\n", pwd);
#endif

	ret = dp_verify((uint8_t *)pwd, strlen(pwd));
	memset(pwd, 0, sizeof(pwd));
	
	return ret;
}

/*
	For some of our packets we run out of buffer space, here we set the pointer to
	a buffer that is big enough
*/
static void increaseEthBuffer(bool increase)
{
	if (increase)
	{
		oldBuffer = net_tx_packet;
		net_tx_packet = networkBuffer + (PKTALIGN - 1);
		net_tx_packet -= (ulong)net_tx_packet % PKTALIGN;
	}
	else
	{
		net_tx_packet = oldBuffer;
	}
}

static int initNetwork(void)
{
	int ret = 0;

	eth_halt();
	eth_set_current();
	ret = eth_init();
	if (ret < 0)
	{
		eth_halt();
		printf(" eth_init failed\n");
		return ret;
	}
	return ret;
}


void dp_set_timeout_handler(ulong iv, thand_f *f)
{
	if (iv == 0) 
	{
		time_handler = (thand_f *)0;
	}
	else 
	{
		time_handler = f;
		time_start = get_timer(0);
		time_delta = iv * CONFIG_SYS_HZ / 1000;
	}
}


static void dns_send(const char *hostname)
{
	struct header *header;
	int n, name_len;
	uchar *p, *pkt;
	const char *s;
	const char *name = hostname;
	enum dns_query_type qtype = DNS_A_RECORD;

	pkt = (uchar *)(net_tx_packet + net_eth_hdr_size() + IP_UDP_HDR_SIZE);
	p = pkt;

	/* Prepare DNS packet header */
	header = (struct header *)pkt;
	header->tid = 1;
	header->flags = htons(0x100);	/* standard query */
	header->nqueries = htons(1);		/* Just one query */
	header->nanswers = 0;
	header->nauth = 0;
	header->nother = 0;

	/* Encode DNS name */
	name_len = strlen(name);
	p = (uchar *)&header->data;	/* For encoding host name into packet */

	do {
		s = strchr(name, '.');
		if (!s)
			s = name + name_len;

		n = s - name;			/* Chunk length */
		*p++ = n;			/* Copy length  */
		memcpy(p, name, n);		/* Copy chunk   */
		p += n;

		if (*s == '.')
			n++;

		name += n;
		name_len -= n;
	} while (*s != '\0');

	*p++ = 0;			/* Mark end of host name */
	*p++ = 0;			/* Some servers require double null */
	*p++ = (unsigned char)qtype;	/* Query Type */

	*p++ = 0;
	*p++ = 1;				/* Class: inet, 0x0001 */

	n = p - pkt;				/* Total packet length */

	dp_dns_our_port = random_port();

	net_send_udp_packet(
		(uchar *)net_bcast_ethaddr,
		net_dns_server,
		DNS_SERVICE_PORT, dp_dns_our_port, n);
}

static void dns_timeout_handler(void)
{
	puts("Timeout\n");
	net_set_state(NETLOOP_FAIL);
}

static void dns_handler(uchar *pkt, unsigned dest, struct in_addr sip,
	unsigned src, unsigned len)
{
	struct header *header;
	const unsigned char *p, *e, *s;
	u16 type;
	int found, stop, dlen;
	char ip_str[22];

	if (dest != dp_dns_our_port)
	{
		printf("Not our port\n");
		return;
	}

	/* We sent one query. We want to have a single answer: */
	header = (struct header *)pkt;
	if (ntohs(header->nqueries) != 1)
	{
		return;
	}

	/* Received 0 answers */
	if (header->nanswers == 0) {
		puts("DNS: host not found\n");
		net_set_state(NETLOOP_SUCCESS);
		return;
	}

	/* Skip host name */
	s = &header->data[0];
	e = pkt + len;
	for (p = s; p < e && *p != '\0'; p++)
	{
		continue;
	}

	/* We sent query class 1, query type 1 */
	if (&p[5] > e || get_unaligned_be16(p + 1) != DNS_A_RECORD) {
		puts("DNS: response was not an A record\n");
		net_set_state(NETLOOP_SUCCESS);
		return;
	}

	/* Go to the first answer section */
	p += 5;

	/* Loop through the answers, we want A type answer */
	for (found = stop = 0; !stop && &p[12] < e; ) {
		/* Skip possible name in CNAME answer */
		if (*p != 0xc0) {
			while (*p && &p[12] < e)
				p++;
			p--;
		}

		type = get_unaligned_be16(p + 2);
		if (type == DNS_CNAME_RECORD) {
			/* CNAME answer. shift to the next section */
			printf("Found canonical name\n");
			dlen = get_unaligned_be16(p + 10);
			printf("dlen = %d\n", dlen);
			p += 12 + dlen;
		}
		else if (type == DNS_A_RECORD) {
			found = 1;
			stop = 1;
		}
		else {
			printf("Unknown type\n");
			stop = 1;
		}
	}

	if (found && &p[12] < e) {
		dlen = get_unaligned_be16(p + 10);
		p += 12;
		memcpy(&dp_ip_addr, p, 4);

		if (p + dlen <= e) {
			ip_to_string(dp_ip_addr, ip_str);
		}
		else {
			puts("server responded with invalid IP number\n");
		}
	}
	net_set_state(NETLOOP_SUCCESS);
}

/* Query a IP address through DNS                                            */ 
static int doDNS(const char *hostname)
{
	int ret = CMD_RET_SUCCESS;

	ret = initNetwork();

	dp_set_timeout_handler(DNS_TIMEOUT, dns_timeout_handler);
	net_set_udp_handler(dns_handler);

restart:
	net_set_state(NETLOOP_CONTINUE);

	dns_send(hostname);
	for (;;)
	{
		/*
		*	Check the ethernet for a new packet.  The ethernet
		*	receive routine will process it.
		*	Most drivers return the most recent packet size, but not
		*	errors that may have happened.
		*/
		eth_rx();
		if (ctrlc()) {
			/* cancel any ARP that may not have completed */

			net_set_udp_handler(NULL);
			net_set_arp_handler(NULL);
			dp_set_timeout_handler(0, NULL);
			eth_halt();
			eth_set_last_protocol(BOOTP);
			puts("\nAbort\n");
			ret = -EINTR;
			goto done;
		}
		/*
		*	Check for a timeout, and run the timeout handler
		*	if we have one.
		*/

		if (time_handler &&
			((get_timer(0) - time_start) > time_delta)) {
			thand_f *x;

			printf("TIMEOUT\n");

			x = time_handler;
			time_handler = (thand_f *)0;
			(*x)();
		}

		if (net_state == NETLOOP_FAIL)
		{
			ret = net_start_again();
		}

		switch (net_state)
		{
		case NETLOOP_RESTART:
			//			net_restarted = 1;
			goto restart;

		case NETLOOP_SUCCESS:
			printf("DNS success\n");
			net_set_udp_handler(NULL);
			net_set_arp_handler(NULL);
			dp_set_timeout_handler(0, NULL);
			eth_halt();
			eth_set_last_protocol(DHCP);
			// what return value to set ???
			ret = 1;
			goto done;

		case NETLOOP_FAIL:
			printf("fail\n");

			net_set_udp_handler(NULL);
			net_set_arp_handler(NULL);
			dp_set_timeout_handler(0, NULL);
			/* Invalidate the last protocol */
			eth_set_last_protocol(BOOTP);
			goto done;

		case NETLOOP_CONTINUE:
			continue;
		}
	}

done:
	return ret;
}



/**
* An incoming packet handler.
* @param pkt    pointer to the application packet
* @param dport  destination UDP port
* @param sip    source IP address
* @param sport  source UDP port
* @param len    packet length
*/
static void challenge_handler(uchar *pkt, unsigned dest, struct in_addr sip,
	unsigned src, unsigned len)
{
	printf("Response from server recieved: \n");

	if (dest != serverPort + 1)
	{
		printf("Not our port!\n");
		return;
	}

	if (src != serverPort)
	{
		printf("Not send from our server port!\n");
		return;
	}

	if (len < 13)
	{
		printf("Response must be at least 13 bytes\n");
		return;
	}

	// Either wait or login
	uint8_t cmdTag = (*(pkt + 4));

	if (RESPONSE_TAG_HASHEDLOGIN_V1 == cmdTag)
	{
		// Size check: Must be 9 + 32
		if (len != 41)
		{
			printf("Hash PIN response must be at least 41 bytes\n");
			return;
		}

		// Set status to HASHPIN_RECEIVED
		statusServer = STATUS_DONE;

		/* Do the login with the received response */
		/* Send hash. */
		uint8_t ins = 0x30;   // PIN login

		struct {
			uint8_t header[5];
			uint8_t length;
			uint8_t hash[DP_HASH_LENGTH];
		} __attribute__((packed)) hash_request = {
			.header = { 0xff, ins, 0x00, 0x00, 0x21 },
			.length = 0x20,
			.hash = { 0 },
		};
		uint8_t hash_response[2];

		memcpy(hash_request.hash, pkt + 9,
			sizeof(hash_request.hash));

		if (dp_cmd((void *)&hash_request, sizeof(hash_request),
			(void *)&hash_response, sizeof(hash_response)) !=
			sizeof(hash_response))
		{
			printf(
				"Failed to do hashed based login to card in SD card slot: unexpected return length\n");
			goto failed;
		}
		if (!(hash_response[0] == 0x90 && hash_response[1] == 0x00))
		{
			printf(
				"Failed to do hashed based login to card in SD card slot: unexpected return code\n");
			goto failed;
		}
		// If we reach here the login succeeded
		// Set status to
	}
	else if (RESPONSE_TAG_WAIT_V1 == cmdTag)
	{
		unsigned long secs = 0;
		secs = (*(pkt + 9) << 24) + (*(pkt + 10) << 16) + (*(pkt + 11) << 8) +
			*(pkt + 12);
		statusServer = STATUS_WAIT;
		end = get_timer(0) + 1000 * secs;
	}

failed:

	validAnswerReceived = true;
}

/*
*	Timeout on UDP challenge
*/
static void dp_challenge_timeout_handler(void)
{
	ulong time_taken = get_timer(dp_start);

	if (time_taken >= time_taken_max)
	{
		puts("\nRetry time exceeded; aborting\n");
		abort = true;
	}
	else
	{
		dp_timeout *= 2;
		if (dp_timeout > 2000)
		{
			dp_timeout = 2000;
		}
		dp_set_timeout_handler(dp_timeout, dp_challenge_timeout_handler);
		challengeTimeout = true;
	}
}


/*
*	Timeout on DHCP request.
*/
static void dp_dhcp_timeout_handler(void)
{
	ulong time_taken = get_timer(dhcp_start);

	if (time_taken >= time_taken_max) 
	{
		puts("\nRetry time exceeded; starting again\n");
		net_set_state(NETLOOP_FAIL);
	}
	else 
	{
		dhcp_timeout *= 2;
		if (dhcp_timeout > 2000)
		{
			dhcp_timeout = 2000;
		}
		dp_set_timeout_handler(dhcp_timeout, dp_dhcp_timeout_handler);
		dp_dhcp_request();
	}
}

static void dhcp_add_id(ulong id)
{
	if (dhcp_num_ids >= ARRAY_SIZE(dhcp_ids)) {
		size_t size = sizeof(dhcp_ids) - sizeof(id);

		memmove(dhcp_ids, &dhcp_ids[1], size);
		dhcp_ids[dhcp_num_ids - 1] = id;
	}
	else {
		dhcp_ids[dhcp_num_ids] = id;
		dhcp_num_ids++;
	}
}

static int bootp_extended(u8 *e)
{
	u8 *start = e;

	*e++ = 99;		/* RFC1048 Magic Cookie */
	*e++ = 130;
	*e++ = 83;
	*e++ = 99;

	*e++ = 53;		/* DHCP Message Type */
	*e++ = 1;
	*e++ = DHCP_DISCOVER;

	*e++ = 57;		/* Maximum DHCP Message Size */
	*e++ = 2;
	*e++ = (576 - 312 + OPT_FIELD_SIZE) >> 16;
	*e++ = (576 - 312 + OPT_FIELD_SIZE) & 0xff;

	*e++ = 6;		/* Domain Name Server */
	*e++ = 4;
	e += 4;

	*e++ = 12;		/* Host name request */
	*e++ = 32;
	e += 32;

	*e++ = 255;		/* End of the list */

	/*
	* If nothing in list, remove it altogether. Some DHCP servers get
	* upset by this minor faux pas and do not respond at all.
	*/
	if (e == start + 3) {
		printf("*** Warning: no DHCP options requested\n");
		e -= 3;
	}
	return e - start;
}

static void store_net_params(struct bootp_hdr *bp)
{
	struct in_addr tmp_ip;

	net_copy_ip(&tmp_ip, &bp->bp_siaddr);
	if (tmp_ip.s_addr != 0)
		net_copy_ip(&net_server_ip, &bp->bp_siaddr);
	memcpy(net_server_ethaddr,
		((struct ethernet_hdr *)net_rx_packet)->et_src, 6);
	if (
		(strlen(bp->bp_file) > 0)) {
		copy_filename(net_boot_file_name, bp->bp_file,
			sizeof(net_boot_file_name));
	}

	/* Propagate to environment:
	* don't delete exising entry when BOOTP / DHCP reply does
	* not contain a new value
	*/
	if (*net_boot_file_name)
		env_set("bootfile", net_boot_file_name);

	net_copy_ip(&net_ip, &bp->bp_yiaddr);
}


static bool bootp_match_id(ulong id)
{
	unsigned int i;

	for (i = 0; i < dhcp_num_ids; i++)
		if (dhcp_ids[i] == id)
			return true;

	return false;
}

static int check_reply_packet(uchar *pkt, unsigned dest, unsigned src,
	unsigned len)
{
	struct bootp_hdr *bp = (struct bootp_hdr *)pkt;
	int retval = 0;

	if (dest != PORT_BOOTPC || src != PORT_BOOTPS)
		retval = -1;
	else if (len < sizeof(struct bootp_hdr) - OPT_FIELD_SIZE)
		retval = -2;
	else if (bp->bp_op != OP_BOOTREPLY)
		retval = -3;
	else if (bp->bp_htype != HWT_ETHER)
		retval = -4;
	else if (bp->bp_hlen != HWL_ETHER)
		retval = -5;
	else if (!bootp_match_id(net_read_u32(&bp->bp_id)))
		retval = -6;
	else if (memcmp(bp->bp_chaddr, net_ethaddr, HWL_ETHER) != 0)
		retval = -7;

	return retval;
}

static int truncate_sz(const char *name, int maxlen, int curlen)
{
	if (curlen >= maxlen) {
		printf("*** WARNING: %s is too long (%d - max: %d)"
			" - truncated\n", name, curlen, maxlen);
		curlen = maxlen - 1;
	}
	return curlen;
}

static void bootp_process_vendor_field(u8 *ext)
{
	int size = *(ext + 1);

	net_boot_file_expected_size_in_blocks = 0;

	switch (*ext) {
		/* Fixed length fields */
	case 1:			/* Subnet mask */
		if (net_netmask.s_addr == 0)
			net_copy_ip(&net_netmask, (struct in_addr *)(ext + 2));
		break;
	case 2:			/* Time offset - Not yet supported */
		break;
		/* Variable length fields */
	case 3:			/* Gateways list */
		if (net_gateway.s_addr == 0)
			net_copy_ip(&net_gateway, (struct in_addr *)(ext + 2));
		break;
	case 4:			/* Time server - Not yet supported */
		break;
	case 5:			/* IEN-116 name server - Not yet supported */
		break;
	case 6:
		if (net_dns_server.s_addr == 0)
			net_copy_ip(&net_dns_server,
			(struct in_addr *)(ext + 2));
		break;
	case 7:			/* Log server - Not yet supported */
		break;
	case 8:			/* Cookie/Quote server - Not yet supported */
		break;
	case 9:			/* LPR server - Not yet supported */
		break;
	case 10:		/* Impress server - Not yet supported */
		break;
	case 11:		/* RPL server - Not yet supported */
		break;
	case 12:		/* Host name */
		if (net_hostname[0] == 0) {
			size = truncate_sz("Host Name",
				sizeof(net_hostname), size);
			memcpy(&net_hostname, ext + 2, size);
			net_hostname[size] = 0;
		}
		break;
	case 13:		/* Boot file size */
		if (size == 2)
			net_boot_file_expected_size_in_blocks =
			ntohs(*(ushort *)(ext + 2));
		else if (size == 4)
			net_boot_file_expected_size_in_blocks =
			ntohl(*(ulong *)(ext + 2));
		break;
	case 14:		/* Merit dump file - Not yet supported */
		break;
	case 15:		/* Domain name - Not yet supported */
		break;
	case 16:		/* Swap server - Not yet supported */
		break;
	case 17:		/* Root path */
		if (net_root_path[0] == 0) {
			size = truncate_sz("Root Path",
				sizeof(net_root_path), size);
			memcpy(&net_root_path, ext + 2, size);
			net_root_path[size] = 0;
		}
		break;
	case 18:		/* Extension path - Not yet supported */
					/*
					* This can be used to send the information of the
					* vendor area in another file that the client can
					* access via TFTP.
					*/
		break;
		/* IP host layer fields */
	case 40:		/* NIS Domain name */
		if (net_nis_domain[0] == 0) {
			size = truncate_sz("NIS Domain Name",
				sizeof(net_nis_domain), size);
			memcpy(&net_nis_domain, ext + 2, size);
			net_nis_domain[size] = 0;
		}
		break;
		/* Application layer fields */
	case 43:		/* Vendor specific info - Not yet supported */
					/*
					* Binary information to exchange specific
					* product information.
					*/
		break;
		/* Reserved (custom) fields (128..254) */
	}
}

static void bootp_process_vendor(u8 *ext, int size)
{
	u8 *end = ext + size;

	while ((ext < end) && (*ext != 0xff)) {
		if (*ext == 0) {
			ext++;
		}
		else {
			u8 *opt = ext;

			ext += ext[1] + 2;
			if (ext <= end)
				bootp_process_vendor_field(opt);
		}
	}

	if (net_dns_server.s_addr)
	{
		printf("net_dns_server	: %pI4\n", &net_dns_server);
	}

	if (net_netmask.s_addr)
		printf("net_netmask     : %pI4\n", &net_netmask);

	if (net_gateway.s_addr)
		printf("net_gateway	    : %pI4\n", &net_gateway);

	if (net_hostname[0])
		printf("net_hostname    : %s\n", net_hostname);
}


/*
*	Handle a DHCP received packet.
*/
static void dp_dhcp_handler(uchar *pkt, unsigned dest, struct in_addr sip,
	unsigned src, unsigned len)
{
	struct bootp_hdr *bp;

	bp = (struct bootp_hdr *)pkt;

	/* Filter out pkts we don't want */
	if (check_reply_packet(pkt, dest, src, len))
		return;

	/*
	*	Got a good DHCP reply.	 Copy the data into our variables.
	*/

	store_net_params(bp);		/* Store net parameters from reply */

	/* Retrieve extended information (we must parse the vendor area) */
	if (net_read_u32((u32 *)&bp->bp_vend[0]) == htonl(VENDOR_MAGIC))
	{
		bootp_process_vendor((uchar *)&bp->bp_vend[4], len);
	}

	dp_set_timeout_handler(0, (thand_f *)0);
	net_set_state(NETLOOP_SUCCESS);
	printf("Our IP address is %pI4 \n",
		 &net_ip);
}

static void dp_dhcp_request(void)
{
	uchar *pkt, *iphdr;
	struct bootp_hdr *bp;
	int extlen, pktlen, iplen;
	int eth_hdr_size;
	u32 bootp_id;
	struct in_addr zero_ip;
	struct in_addr bcast_ip;

	printf("DHCP broadcast %d\n", ++dhcp_try);

	pkt = net_tx_packet;
	memset((void *)pkt, 0, PKTSIZE);

	eth_hdr_size = net_set_ether(pkt, net_bcast_ethaddr, PROT_IP);
	pkt += eth_hdr_size;

	/*
	* Next line results in incorrect packet size being transmitted,
	* resulting in errors in some DHCP servers, reporting missing bytes.
	* Size must be set in packet header after extension length has been
	* determined.
	* C. Hallinan, DS4.COM, Inc.
	*/
	/* net_set_udp_header(pkt, 0xFFFFFFFFL, PORT_BOOTPS, PORT_BOOTPC,
	sizeof (struct bootp_hdr)); */
	iphdr = pkt;	/* We need this later for net_set_udp_header() */
	pkt += IP_UDP_HDR_SIZE;

	bp = (struct bootp_hdr *)pkt;
	bp->bp_op = OP_BOOTREQUEST;
	bp->bp_htype = HWT_ETHER;
	bp->bp_hlen = HWL_ETHER;
	bp->bp_hops = 0;
	/*
	* according to RFC1542, should be 0 on first request, secs since
	* first request otherwise
	*/
	bp->bp_secs = htons(get_timer(dhcp_start) / 1000);
	zero_ip.s_addr = 0;

	net_write_ip(&bp->bp_ciaddr, zero_ip);
	net_write_ip(&bp->bp_yiaddr, zero_ip);
	net_write_ip(&bp->bp_siaddr, zero_ip);
	net_write_ip(&bp->bp_giaddr, zero_ip);

	memcpy(bp->bp_chaddr, net_ethaddr, 6);

	copy_filename(bp->bp_file, net_boot_file_name, sizeof(bp->bp_file));
	extlen = bootp_extended((u8 *)bp->bp_vend);

	/*
	*	Bootp ID is the lower 4 bytes of our ethernet address
	*	plus the current time in ms.
	*/
	bootp_id = ((u32)net_ethaddr[2] << 24)
		| ((u32)net_ethaddr[3] << 16)
		| ((u32)net_ethaddr[4] << 8)
		| (u32)net_ethaddr[5];
	bootp_id += get_timer(0);
	bootp_id = htonl(bootp_id);
	dhcp_add_id(bootp_id);
	net_copy_u32(&bp->bp_id, &bootp_id);

	/*
	* Calculate proper packet lengths taking into account the
	* variable size of the options field
	*/
	iplen = BOOTP_HDR_SIZE - OPT_FIELD_SIZE + extlen;
	pktlen = eth_hdr_size + IP_UDP_HDR_SIZE + iplen;
	bcast_ip.s_addr = 0xFFFFFFFFL;

	net_set_udp_header(iphdr, bcast_ip, PORT_BOOTPS, PORT_BOOTPC, iplen);
	dp_set_timeout_handler(dhcp_timeout, dp_dhcp_timeout_handler);
	net_set_udp_handler(dp_dhcp_handler);
	net_send_packet(net_tx_packet, pktlen);
}


static int getDHCPLease(void)
{
	int ret = -EINVAL;
	ret = initNetwork();

restart:
	net_set_state(NETLOOP_CONTINUE);
	if (eth_get_dev())
	{
		memcpy(net_ethaddr, eth_get_ethaddr(), 6);
	}

	dhcp_num_ids		= 0;
	dhcp_try			= 0;
	dhcp_start			= get_timer(0);
	dhcp_timeout		= 200;
	net_ip.s_addr = 0;

	dp_dhcp_request();
	for (;;)
	{
		/*
		*	Check the ethernet for a new packet.  The ethernet
		*	receive routine will process it.
		*	Most drivers return the most recent packet size, but not
		*	errors that may have happened.
		*/
		eth_rx();
		if (ctrlc()) {
			/* cancel any ARP that may not have completed */

			net_set_udp_handler(NULL);
			net_set_arp_handler(NULL);
			dp_set_timeout_handler(0, NULL);
			eth_halt();
			eth_set_last_protocol(BOOTP);
			puts("\nAbort\n");
			ret = -EINTR;
			goto done;
		}
		/*
		*	Check for a timeout, and run the timeout handler
		*	if we have one.
		*/

		if (time_handler &&
			((get_timer(0) - time_start) > time_delta)) {
			thand_f *x;

			x = time_handler;
			time_handler = (thand_f *)0;
			(*x)();
		}

		if (net_state == NETLOOP_FAIL)
		{
			ret = net_start_again();
		}

		switch (net_state) 
		{
		case NETLOOP_RESTART:
			goto restart;

		case NETLOOP_SUCCESS:
			printf("DHCP success\n");

			net_set_udp_handler(NULL);
			net_set_arp_handler(NULL);
			dp_set_timeout_handler(0, NULL);
			eth_halt();
			eth_set_last_protocol(DHCP);
			// what return value to set ???
			ret = 1;
			goto done;

		case NETLOOP_FAIL:
			net_set_udp_handler(NULL);
			net_set_arp_handler(NULL);
			dp_set_timeout_handler(0, NULL);
			/* Invalidate the last protocol */
			eth_set_last_protocol(BOOTP);
			goto done;

		case NETLOOP_CONTINUE:
			continue;
		}
	}

done:
	return ret;
}

static int getControllerId(uint8_t *id, ssize_t *len)
{
	int ret = CMD_RET_SUCCESS;
	ssize_t bytesToCopy = 32;

	static const uint8_t controller_request[5] =
	{ 0xff, 0x70, 0x06, 0x00, 0x00 };
	struct {
		uint8_t id[64];
		uint16_t code;
	} __attribute__((packed)) controller_response;

	ssize_t length = dp_cmd((void *)&controller_request,
		sizeof(controller_request),
		(void *)&controller_response, sizeof(controller_response));

	/* Controller ID must be between 8 and 64 bytes in size                  */
	if ((length < 10) || (length > 66))
	{
		printf(
			"Failed to retrieve controller id from card in SD card slot: unexpected return length <%d>\n",
			length);

		ret = CMD_RET_FAILURE;
		goto failed;
	}

	if (!(controller_response.id[length - 2] == 0x90 && 
								controller_response.id[length - 1] == 0x00))
	{
		printf(
			"Failed to retrieve controller id  from card in SD card slot: unexpected return code\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}
	
	// When we reach here command has succeeeded
	length -= 2;
	*len = length;

	// This is weird CFE pads the 0 at the end, server wants it at the 
	// start and only 32 byte data
	memset(id, 0, 32);
	if (length < bytesToCopy)
	{
		bytesToCopy = length;
	}

	memcpy(id + 32 - bytesToCopy, controller_response.id, bytesToCopy);

failed:
	return ret;
}


static int  sendNetPolicyCommand(uint8_t command)
{
	uint8_t packetCmdHashLogin[sizeof(COMMAND_HEADER) 
										+ LENGTH_HASHEDLOGIN_PAYLOAD_V1];
	uint8_t packetCmdPoll[sizeof(COMMAND_HEADER) + LENGTH_POLL_PAYLOAD_V1];

	int ret = CMD_RET_SUCCESS;
	
	/* Retrieve controller id from DP card */
	ssize_t length = 0;
	uint8_t id[32];
	ret = getControllerId(id, &length);
	if (CMD_RET_SUCCESS != ret)
	{
		goto failed;
	}

	uint8_t challenge[DP_CHALLENGE_LENGTH];
	ret = getLoginChallenge(challenge);
	if (CMD_RET_SUCCESS != ret)
	{
		goto failed;
	}

	switch (command)
	{
		case COMMAND_TAG_HASHEDLOGIN_V1:
			increaseEthBuffer(true);
			memset(packetCmdHashLogin, 0, sizeof(packetCmdHashLogin));
			/* Header */
			memcpy(packetCmdHashLogin, &headerHashedLogin, 
											sizeof(headerHashedLogin));
			/* ID */
			memcpy(packetCmdHashLogin + sizeof(headerHashedLogin), id, 
																	sizeof(id));
			/* Challenge */
			memcpy(packetCmdHashLogin + sizeof(headerHashedLogin) 
				+ sizeof(id) + 32,
				challenge, sizeof(challenge));
			memcpy((char *)net_tx_packet + net_eth_hdr_size() + IP_UDP_HDR_SIZE,
				packetCmdHashLogin, sizeof(packetCmdHashLogin));
			break;
		case COMMAND_TAG_POLL_V1:
			memset(packetCmdPoll, 0, sizeof(packetCmdPoll));
			memcpy(packetCmdPoll, &headerPoll, sizeof(headerPoll));
			memcpy(packetCmdPoll + sizeof(headerPoll), id, sizeof(id));
			memcpy((char *)net_tx_packet + net_eth_hdr_size() + IP_UDP_HDR_SIZE,
				packetCmdPoll, sizeof(packetCmdPoll));
			break;
		default:
			break;
	}

	net_set_udp_handler(challenge_handler);

	dp_timeout	= 250;
	dp_start	= get_timer(0);

	initNetwork();

resend:
	challengeTimeout = false;
	dp_set_timeout_handler(dp_timeout, dp_challenge_timeout_handler);
	net_send_udp_packet(
			(uchar *)net_null_ethaddr,
			dp_ip_addr, serverPort,
			serverPort + 1,
			sizeof(packetCmdHashLogin));
	
	printf("Waiting for reply from server\n");
	validAnswerReceived = false;
	for (;;)
	{
		if (validAnswerReceived)
		{
			dp_set_timeout_handler(0, NULL);
			eth_halt();
			break;
		}

		if (abort)
		{
			return CMD_RET_FAILURE;
		}

		if (challengeTimeout)
		{
			goto resend;
		}

		eth_rx();
		if (ctrlc()) {

			eth_halt();
			/* Invalidate the last protocol */
			puts("\nAbort\n");
			goto failed;
		}


		/*
		*	Check for a timeout, and run the timeout handler
		*	if we have one.
		*/
		if (time_handler &&
			((get_timer(0) - time_start) > time_delta)) {
			thand_f *x;

			x = time_handler;
			time_handler = (thand_f *)0;
			(*x)();
		}
	}

failed:
	increaseEthBuffer(false);
	return ret;

}

// TODO: have a sepaarte function initNetwork

static int  doNetDongle(void)
{
	int ret;

	net_init();

	// only do this if neccessary
	getDHCPLease();
	if (needDNSResolution)
	{
		// Configuration data does not include an ip address but an
		// arbitrary string, try to resolve it

		// First make sure we can reach our DNS server
		net_ping_ip = net_dns_server;
		if (net_ping_ip.s_addr == 0)
		{
			printf("No valid addr for ping given\n");
			return CMD_RET_FAILURE;
		}

		ret = net_loop(PING);
		if (0 != ret)
		{
			printf("ping to %pI4 failed: error: <%d>\n", &net_ping_ip, ret);
			return CMD_RET_FAILURE;
		}
		printf("DNS server %pI4 is alive\n", &net_ping_ip);
		doDNS(serverAddr);
	}
	else
	{
		dp_ip_addr = string_to_ip(serverAddr);
	}
	printf("IP addr for %s is %pI4 \n", serverAddr, &dp_ip_addr);

	net_ping_ip = dp_ip_addr;
	if (net_ping_ip.s_addr == 0)
	{
		printf("No valid addr for ping given\n");
		return CMD_RET_FAILURE;
	}

	ret = net_loop(PING);
	if (0 != ret)
	{
		printf("failed: error: <%d>\n", ret);
		return CMD_RET_FAILURE;
	}
	printf("DP login server %pI4  is alive\n", &dp_ip_addr);

	ret = sendNetPolicyCommand(COMMAND_TAG_HASHEDLOGIN_V1);

	// As long as we receive the wait flag we poll the server for a result
	while (statusServer == STATUS_WAIT)
	{
		unsigned long now = get_timer(0);
		while (now < end)
		{
			now = get_timer(0);
		}
		ret = sendNetPolicyCommand(COMMAND_TAG_POLL_V1);
	}
	return ret;
}

static int doDongle(void)
{
	int ret = CMD_RET_SUCCESS;

	/* Start with MMC selected */
	useUSBDevice = false;

	/* Retrieve challenge from DP card */
	static const uint8_t challenge_request[5] =
	{ 0xff, 0x70, 0x05, 0x00, 0x00 };
	struct {
		uint8_t challenge[DP_CHALLENGE_LENGTH];
		uint16_t code;
	} __attribute__((packed)) challenge_response;

	if (dp_cmd((void *)&challenge_request, sizeof(challenge_request),
		(void *)&challenge_response, sizeof(challenge_response)) !=
		sizeof(challenge_response))
	{
		printf(
			"Failed to retrieve challenge from card in SD card slot: unexpected return length\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}

	if (ntohs(challenge_response.code) != 0x9000)
	{
		printf(
			"Failed to retrieve challenge from card in SD card slot: unexpected return code\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}

	/* Retrieve authenticity secret from        CCID                            */
	useUSBDevice = true;

	initCCID();
	/* If we have an CCID device we retrieve the authenticity secret from there */
	if (NULL != udevCCID)
	{
		setTarget(DEVICE_CCID);
	}
	else
	{
		ret = CMD_RET_FAILURE;
		if ((0 == vendorIdFromPolicy) || (0 == productIdFromPolicy))
		{
			printf("No CCID device found to provide an unlock code\n");
		}
		else
		{
			printf(
"No CCID device with vendor id 0x%x device id 0x%x found to provide an unlock code\n",
				vendorIdFromPolicy, productIdFromPolicy);
		}
		goto failed;
	}

	uint8_t authenticity_request[37];
	uint8_t *rq = authenticity_request;

	*rq++ = 0xff;
	*rq++ = 0xf2;
	*rq++ = 0x00;
	*rq++ = 0x00;
	*rq++ = 0x20;

	memcpy(rq, challenge_response.challenge,
		sizeof(challenge_response.challenge));

	struct {
		uint8_t response[DP_AUTHENTICITY_LENGTH];
		uint16_t code;
	} __attribute__((packed)) authenticity_response;

	if (dp_cmd((void *)&authenticity_request,
		sizeof(authenticity_request),
		(void *)&authenticity_response, sizeof(authenticity_response)) !=
		sizeof(authenticity_response)) {
		printf(
"Failed to retrieve hashed login from USB masstorage device 0: unexpected return length\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}
	if (ntohs(authenticity_response.code) != 0x9000) {
		printf(
"Failed to retrieve hashed login from USB masstorage device 0: unexpected return code\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}

	// Switch back to MMC
	useUSBDevice = false;
	setTarget(DEVICE_ASSD);

	/* Send hash. */
	uint8_t ins = 0x30;   // PIN login

	struct {
		uint8_t header[5];
		uint8_t length;
		uint8_t hash[DP_HASH_LENGTH];
	} __attribute__((packed)) hash_request = {
		.header = { 0xff, ins, 0x00, 0x00, 0x21 },
		.length = 0x20,
		.hash = { 0 },
	};
	uint8_t hash_response[2];

	memcpy(hash_request.hash, authenticity_response.response,
		sizeof(hash_request.hash));

	if (dp_cmd((void *)&hash_request, sizeof(hash_request),
		(void *)&hash_response, sizeof(hash_response)) !=
		sizeof(hash_response))
	{
		printf(
"Failed to do hashed based login to card in SD card slot: unexpected return length\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}
	if (!(hash_response[0] == 0x90 && hash_response[1] == 0x00))
	{
		printf(
"Failed to do hashed based login to card in SD card slot: unexpected return code\n");
		ret = CMD_RET_FAILURE;
		goto failed;
	}

failed:
	/* We must leave this function with USB deselected */
	useUSBDevice = false;
	return ret;
}


static int executePolicy(const int policy, const bool isHashed)
{
	int ret = CMD_RET_SUCCESS;
	struct blk_desc *dummyDesc = NULL;

	if (policy > MAXPOLICY)
	{
		return CMD_RET_FAILURE;
	}

	switch (policy)
	{
		case 0:
			// Nothing to do
			break;
		case 1:
			if (isHashed)
			{
				ret = doHashedPIN();
			}
			else
			{
				ret = doPIN();
			}
			break;
		case 2:
			// init USB now, it is stupid but needed twice
			usbReset();
			usbReset();
			blk_print_device_num(IF_TYPE_USB, 0);
			dummyDesc = blk_get_devnum_by_type(IF_TYPE_USB, 0);
			if (NULL != dummyDesc)
			{
				if (1 == dummyDesc->removable)
				{
					if (DEV_TYPE_HARDDISK == (dummyDesc->type & 0x1F))
					{
						ret = doDongle();
					}
				}
			}

			break;
		case 3:
			if (isHashed)
			{
				ret = doNetDongle();
			}
			else
			{
				ret = doPIN();
			}
			break;
		default:
			return CMD_RET_FAILURE;
	}
	return ret;
}

#if 0
// If card is not activated, ask to activate and select a policy
static int activateAndVerify(int *state)
{
	char pwd[DP_PASSWORD_MAX_LENGTH + 1];
	char soPwd[DP_PASSWORD_MAX_LENGTH + 1];
	int ret = CMD_RET_SUCCESS;
	bool doActivate = false;
	int res = 0;
	int count = 0;

	// Ask to activate
	printf("Do you want to activate the DP card now?\n");
	printf("Press [y] or [n] within five seconds ");

	uint64_t lasttick = (get_ticks() + (uint64_t)(5)* get_tbclk());
	while ((110 != res) && (121 != res) && (lasttick > get_ticks()))
	{
		if (tstc())
		{
			res = getc();
		}
		if ((110 != res) && (121 != res) && (0 != res))
		{
			printf("\nEnter y or n ");
			res = 0;
		}
		else if (121 == res)
		{
			printf("y");
			doActivate = true;
		}
		else if (110 == res)
		{
			// User presses n
			printf("n\n");
		}
	}
	if (doActivate)
	{
		printf("\n Enter your PIN: ");

		memset(pwd, 0, sizeof(pwd));
		do
		{
			res = getc();
			if (res != 13)
			{
				pwd[count] = (char)res;
				printf("*");
			}
			count++;
		} while (res != 13 && res >= 0 && count < DP_PASSWORD_MAX_LENGTH);
		printf("\n");

#ifdef PRINT_VERBOSE
		printf("Your PIN is: <%s>\n", pwd);
#endif

		printf("\n Enter your SO PIN: ");

		memset(soPwd, 0, sizeof(soPwd));
		count = 0;
		do
		{
			res = getc();
			if (res != 13)
			{
				soPwd[count] = (char)res;
				printf("*");
			}
			count++;
		} while (res != 13 && res >= 0 && count < DP_PASSWORD_MAX_LENGTH);
		printf("\n");

#ifdef PRINT_VERBOSE
		printf("Your SO PIN is: <%s>\n", soPwd);
#endif

		res = dp_activate((uint8_t *)pwd, strlen(pwd),
			(uint8_t *)soPwd, strlen(soPwd), 10);
		if (res != CMD_RET_SUCCESS)
		{
			printf("Failed to activate card!\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}

#ifdef DO_CONFIG

		// Write default policy
		res = write_defaultPolicy();
		if (res != CMD_RET_SUCCESS)
		{
			printf("Failed to write policy!\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}
#endif

		res = dp_verify((uint8_t *)pwd, strlen(pwd));
		memset(soPwd, 0, sizeof(soPwd));
		memset(pwd, 0, sizeof(pwd));

		if (res != CMD_RET_SUCCESS)
		{
			printf("Verify failed!\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}
		*state = 1;
	}
	else
	{
		// User does not want to activate
		// Not a failure, just move on
		*state = 0;
		ret = CMD_RET_SUCCESS;
	}

failed:
	return ret;
}
#endif

static int readPolicy(void)
{
	int policy				= NO_POLICY;
	uint16_t expectedSize	= 0;
	bool isIPAddr			= false;

	// Read block 0 of random NVRAM
	// Get the size
	int ret = dp_readConfigSize(&expectedSize);
	if (ret != CMD_RET_SUCCESS)
	{
		printf("dp_readConfigSize failed\n");
		goto failed;
	}

	if (expectedSize < 3)
	{
		goto failed;
	}

	uint8_t data[512];
	memset(data, 0, sizeof(data));

	ret = dp_readConfig(data, expectedSize);
	if (ret != CMD_RET_SUCCESS)
	{
		printf("dp_readConfig failed\n");
		goto failed;
	}

	if (expectedSize > 2)
	{
		expectedSize -= 2;
	}

	// Look into first 3 bytes
	// if usb => POLICY_USB
	// if pin => POLICY_PIN
	if ((data[0] < 0x40) || (data[1] < 0x40) || (data[2] < 0x40))
	{
		goto failed;
	}

	// Poor mans toupper
	for (int i = 0; i < 3; i++)
	{
		if (data[i] > 0x5a) 
		{
			data[i] -= 0x20;
		}
	}

	// "USB"
	if ((data[0] == 0x55) && (data[1] == 0x53) && (data[2] == 0x42))
	{
		policy = USB_POLICY;
		/* Check for usb vendor id product id                                   */
		/* USB#1234#abcd                                                        */
		/* length must be 13 characters, data[3] and data[8] must be 0x23       */
		if (11 == expectedSize)
		{
			if ((data[3] == 0x23) && (data[8] == 0x23))
			{
				data[8]		= 0;
				data[13]	= 0;

				vendorIdFromPolicy	= 
							simple_strtoul((const char *)&data[4], NULL, 16);
				productIdFromPolicy = 
							simple_strtoul((const char *)&data[9], NULL, 16);
				
				// Make sure both are valid
				if (0 == vendorIdFromPolicy)
				{
					productIdFromPolicy = 0;
				}
				if (0 == productIdFromPolicy)
				{
					vendorIdFromPolicy = 0;
				}
			}
		}
		printf("Found USB_POLICY VendorId <0x%x> ProductId <0x%x\n",
									vendorIdFromPolicy, productIdFromPolicy);
	}
	// "PIN"
	else if  ((data[0] == 0x50) && (data[1] == 0x49) && (data[2] == 0x4e))
	{
		printf("Found PIN_POLICY\n");
		policy = PIN_POLICY;
	}
	// "NET#"
	else if ((data[0] == 0x4E) &&   // N
			(data[1] == 0x45) &&	// E
			(data[2] == 0x54) &&	// T
			(data[3] == 0x23))		// #
	{
		int hashPosition = 0;

		// We need at least x.x.x.x#y, ie. 9 additional characters
		if (expectedSize < 13)
		{
			printf("Found NET# expectedSize < 13\n");
			goto failed;
		}

		// Find the next '#', must not be the last one
		for (int i = 5; i < expectedSize - 1; i++)
		{
			if (data[i] == 0x23)
			{
				hashPosition = i;
				break;
			}
		}
		if (0 == hashPosition)
		{
			// # delimiter not found, no NET policy
			printf("0 == hashPosition\n");
			goto failed;
		}
		// all chars behind # must be digits, max 5
		if ((expectedSize - hashPosition) > 5)
		{
			// too many characters
			printf("expectedSize <%d> - hashPosition <%d> > 5\n", expectedSize,
				hashPosition);
			goto failed;
		}
		for (int i = hashPosition + 1; i < expectedSize; i++)
		{
			if ((data[i] < 0x30) || (data[i] > 0x39))
			{
				// not a digit
				printf("not a digit\n");
				goto failed;
			}
		}
		// convert (data + hashPosition + 1) to a integer
		serverPort = simple_strtoul((const char *)(data + hashPosition + 1), 
																	NULL, 10);
		if (serverPort > 65535)
		{
			// not a port
			printf("not a port\n");

			goto failed;
		}

		memset(serverAddr, 0, sizeof(serverAddr));
		memcpy(serverAddr, data + 4, hashPosition - 4);

		// IP addr must have exactly 3 dots
		int numDots	= 0;
		bool wasDot	= true;
		int digitCount	= 0;
		for (int i = 0; i < hashPosition - 4; i++)
		{
			if (wasDot)
			{
				if ((serverAddr[i] < 0x30) || (serverAddr[i] > 0x39))
				{
					// not a digit
					printf("not a digit in Addr <%s> \n", serverAddr);

					goto needDNS;
				}
				digitCount++;
				wasDot = false;
			}
			else
			{
				if ((serverAddr[i] < 0x30) || (serverAddr[i] > 0x39) )
				{
					// not a digit and not a dot
					if (serverAddr[i] != 0x2e)
					{
						printf("not a digit not a dot in Addr <%s>\n",
							serverAddr);
						goto needDNS;
					}
					// It is a dot
					wasDot = true;
					digitCount = 0;
					numDots++;
				}
				else
				{
					digitCount++;
				}
			}
			if (digitCount > 3)
			{
				printf("digit execeed\n");

				goto needDNS;
			}
			if (numDots > 3)
			{
				printf("dot execeed\n");

				goto needDNS;
			}
		}
		if (numDots != 3)
		{
			goto needDNS;
		}
		isIPAddr = true;
needDNS:
		if (!isIPAddr)
		{
			needDNSResolution = true;
		}
		policy = NET_POLICY;
		printf("Found NET policy %s:%d \n", serverAddr, serverPort);
	}

failed:
	return policy;
}

static int checkPolicy(const int policy, const int state)
{
	int ret = CMD_RET_SUCCESS;

	// POLICY_USB needs an USB device
	printf("Policy is %d\n", policy);
	return ret;
}

static int readAndExecutePolicy(const int state, const bool isHashed)
{
	int ret = CMD_RET_SUCCESS;
	if (0 != state)
	{
#ifdef DO_CONFIG
		// Read the policy
		uint16_t policySize = 0;
		uint8_t  policyData[256];

		ret = dp_sizeConfig(&policySize);
		if (ret != CMD_RET_SUCCESS)
		{
			printf("Failed to read policy size!\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}

		ret = dp_readConfig(policyData, policySize);
		if (ret != CMD_RET_SUCCESS)
		{
			printf("Failed to read policy\n");
			ret = CMD_RET_FAILURE;
			goto failed;

		}

		// Evaluate policy
		// Policy size must be 8 (magic data) + 1 (version) + 1 (policy)
		if ((sizeof(MAGIC_CONF) + 2) != policySize)
		{
			printf("wrong policySize: found %d, expected 10\n", policySize);
			ret = CMD_RET_FAILURE;
			goto failed;
		}

		// First eight bytes must match MAGIC_CONF
		const bool hasMagicData = checkMagicData(policyData);
		if (!hasMagicData)
		{
			printf("No U-boot DP configuration\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}

		int policyVersion = *(policyData + sizeof(MAGIC_CONF));
		if (policyVersion > MAXPOLICY)
		{
			printf("Unsopported U-boot policy version\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}

		int policy = *(policyData + sizeof(MAGIC_CONF) + 1);
#else
		const int policy = readPolicy();

		ret = checkPolicy( policy, state );
		if (ret != CMD_RET_SUCCESS)
		{
			printf("checkPolicy failed!\n");
			goto failed;
		}

#endif
		ret = executePolicy(policy, isHashed);
		if (ret != CMD_RET_SUCCESS)
		{
			printf("executePolicy failed\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}
	}
failed:
	return ret;
}

static int do_dp_test(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	uint8_t mode;
	uint8_t state;
	int8_t counter;
	int8_t so_counter;
	uint32_t reset_counter;
	uint32_t cdrom;
	uint8_t flags;
	uint16_t status;
	STATUS_RESPONSE response;
	int res = CMD_RET_SUCCESS;
	unsigned long limit = 10000;

#ifdef CHECK_STACK
	int x = 0;
	top_of_stack = (uint32_t)&x;
#endif 

	if (argc == 2)
	{
		limit = simple_strtoul(argv[1], NULL, 10);
	}

	/* This test fails at about 9000 iterations and card will need a reset      */
	/* TODO: Investigate root cause                                             */

	for (unsigned long i = 0; i < limit; i++)
	{
#if 1
		if (0 == (i % 250))
		{
			printf(
				"Testing ASSD communication: loop %ld of %ld \n", i, limit);
#ifdef CHECK_STACK
				SizeOfStack();
#endif 

		}
#endif

		res = dp_status(&response);
		if (CMD_RET_SUCCESS != res)
		{
			printf("Test failed in dp_status index is:%ld\n", i);
			return res;
		}
		if (0 == i)
		{
			mode = response.mode;
			state = response.state;
			counter = response.counter;
			so_counter = response.so_counter;
			reset_counter = response.reset_counter;
			cdrom = response.cdrom;
			flags = response.flags;
			status = response.status;
		}
		else
		{
			if (mode != response.mode)
			{
				res = CMD_RET_FAILURE;
			}
			if (state != response.state)
			{
				res = CMD_RET_FAILURE;
			}

			if (counter != response.counter)
			{
				res = CMD_RET_FAILURE;
			}

			if (so_counter != response.so_counter)
			{
				res = CMD_RET_FAILURE;
			}
			if (reset_counter != response.reset_counter)
			{
				res = CMD_RET_FAILURE;
			}

			if (cdrom != response.cdrom)
			{
				res = CMD_RET_FAILURE;
			}

			if (flags != response.flags)
			{
				res = CMD_RET_FAILURE;
			}

			if (status != response.status)
			{
				res = CMD_RET_FAILURE;
			}
			if (res == CMD_RET_FAILURE)
			{
				printf("Test failed (invalid result read) index is:%ld\n", i);
				return res;
			}
		}
	}

	return res;
}

static int do_dp_play(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	bool hasCCIDDevice = initCCID();
	STATUS_RESPONSE response;
	int res = CMD_RET_SUCCESS;

	if (hasCCIDDevice)
	{
		unsigned char cmd[] = { 0xff,  0x70,  0x0,  0x0,  0x0 };
		unsigned char rx_buffer[512];
		unsigned int rx_length = sizeof(rx_buffer);

		printf("Found a CCID device\n");

		printf(" udevCCID->devnum %d \n", udevCCID->devnum);
		if (strlen(udevCCID->mf) || strlen(udevCCID->prod))
			printf(" Manufacturer; %s Product: %s \n",
				udevCCID->mf, udevCCID->prod);

		printf(" Vendor Id: 0x%x Device Id 0x%x\n",
			(unsigned int)udevCCID->descriptor.idVendor,
			(unsigned int)udevCCID->descriptor.idProduct);
		
		printf("epDescIn: Endpoint %d %s ", epDescIn->bEndpointAddress & 0xf,
			(epDescIn->bEndpointAddress & 0x80) ? "In" : "Out");
		printf(" MaxPacket %d", get_unaligned(&epDescIn->wMaxPacketSize));
		printf("\n");
		printf("epDescOut: Endpoint %d %s ", epDescOut->bEndpointAddress & 0xf,
			(epDescOut->bEndpointAddress & 0x80) ? "In" : "Out");
		printf(" MaxPacket %d", get_unaligned(&epDescOut->wMaxPacketSize));
		printf("\n");

		printf("Retrieving status from CCID device:\n");

		cmdCCID(
			sizeof(cmd), cmd, &rx_length,
			 rx_buffer);

		printf("\n rx_length 0x%x \n", rx_length);

		for (int i = 0; i < 32; i++)
		{
			printf(" 0x%x ", rx_buffer[i]);
			
		}
		printf("\n");
	}
	else
	{
		printf("No CCID device found\n");
	}

	setTarget(DEVICE_CCID);
	res = dp_status(&response);

	printf("CCID Device:\n");
	printf("  Card Flash Encryption Mode: 0x%02x\n"
		"  System state: 0x%02x\n"
		"  Password retry count: %d\n"
		"  SO Password retry count: %d\n"
		"  Number of Resets: %u\n"
		"  CD-ROM address: 0x%08x\n"
		"  Extended Security flags: 0x%02x\n",
		response.mode, response.state,
		response.counter, response.so_counter,
		ntohl(response.reset_counter),
		ntohl(response.cdrom),
		response.flags);
	setTarget(DEVICE_ASSD);
	res = dp_status(&response);
	printf("\nASSD Device:\n");

	printf("  Card Flash Encryption Mode: 0x%02x\n"
		"  System state: 0x%02x\n"
		"  Password retry count: %d\n"
		"  SO Password retry count: %d\n"
		"  Number of Resets: %u\n"
		"  CD-ROM address: 0x%08x\n"
		"  Extended Security flags: 0x%02x\n",
		response.mode, response.state,
		response.counter, response.so_counter,
		ntohl(response.reset_counter),
		ntohl(response.cdrom),
		response.flags);

	return res;
}

/**
	Implementation of methods to send and Receive using ASSD
*/
static int CMD13(void)
{
	struct mmc_cmd cmd;
	struct mmc *mmc;

	mmc = init_mmc_device(mmc_device, false);
	
	cmd.cmdidx = MMC_CMD_SEND_STATUS;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = mmc->rca << 16;
	int err = mmc_send_cmd(mmc, &cmd, NULL);


#ifdef PRINT_ASSD
	 printf("CMD13 Sent\n");
	 mmmc_trace_ASSD(mmc, &cmd,
					 err);
#endif

	if (err != 0)
	{
		return -1;
	}

	return cmd.response[0];

}

int transmitAPDU(uint8_t *APDU)
{

	int err = CMD_RET_SUCCESS;
	struct mmc_cmd cmd;
	struct mmc *mmc;
	struct mmc_data data;
	uint8_t res[512] = {}; // buffer to receive the response

	// command argument for switching to ASSD
	int arg_cmd6 = 0 | (uint32_t) (1 & 1) << 31 | (uint32_t) (0xf & 0xF) << 20
			| (uint32_t) (0xf & 0xF) << 16 | (uint32_t) (0xf & 0xF) << 12
			| (uint32_t) (0xf & 0xF) << 8 | (uint32_t) (0x4 & 0xF) << 4
			| (uint32_t) (0xf & 0xF);
	
	
	mmc = init_mmc_device(mmc_device, false); // get mmc device
	
	
	// First Send CMD6 
	
	cmd.cmdidx = 6;
	cmd.resp_type = MMC_RSP_R1;
	cmd.cmdarg = arg_cmd6;

	data.blocksize = 64;
	data.dest = (char *)res;
	data.blocks = 1;
	data.flags = MMC_DATA_READ;


	err = mmc_send_cmd(mmc, &cmd, &data);

#ifdef PRINT_ASSD
	 mmmc_trace_ASSD(mmc, &cmd,
					 err);
	printf("CMD 6 response\n");
	print_buffer(0x0000000, res, 1, 64, 0);
#endif

	if (err !=0)
	{
		printf("ASSD command unsuccessfull");
		return err;
	}

#ifdef PRINT_ASSD
	printf("Send CMD13 after CMD6\n");
#endif	
	
	int CMD13Res;

	CMD13Res = CMD13() ;

	if  (CMD13Res < 0 || CMD13Res !=0x0900)
	{
			printf("ASSD command unsuccessfull");
			err = -2;
			
	}	

#ifdef PRINT_ASSD
	printf("Byte 11 %x :  Byte 16 %x \n",res[11] , (res[16] & 0xF0) >>4);
#endif

	// check if card supports ASSD and is switched on 
	if (res[11] != 0x13 && (((res[16] & 0xF0) >>4) != 0x4 ||
												((res[16] & 0xF0) >>4) != 0x2))
	{
		return -1;
	}

#ifdef PRINT_ASSD
	printf("APDU to send\n");
	print_buffer(0x0000000, APDU, 1, 96, 0);
#endif
			
	memset(&cmd, 0, sizeof(cmd));

	// Write APDU 
	cmd.cmdidx = 35;
	cmd.cmdarg = 1;
	cmd.resp_type = MMC_RSP_R1;

	data.src = (char *)APDU; // 
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_WRITE;

	err = mmc_send_cmd(mmc, &cmd, &data);
#ifdef PRINT_ASSD
	mmmc_trace_ASSD(mmc, &cmd,err);
#endif	
	if (err !=0)
	{
		printf("ASSD command unsuccessfull");
		return err;
	}
#ifdef PRINT_ASSD
	printf("send CMD13 after CMD35\n");
#endif	
	
	CMD13Res = CMD13() ;

	if  (CMD13Res < 0 || CMD13Res !=0x0900)
	{
			printf("ASSD command unsuccessfull");
			err = -2;
			
	}	

	// read until response ready
	memset(&cmd, 0, sizeof(cmd));
	cmd.cmdidx = 36;
	cmd.cmdarg = 0;
	cmd.resp_type = MMC_RSP_R1;

	data.dest =(char *) res;
	data.blocks = 1;
	data.blocksize = 32;
	data.flags = MMC_DATA_READ;
	int  Attempts = DP_CMD_ATTEMPTS;
	do{
		err = mmc_send_cmd(mmc, &cmd, &data);
		if (err != 0) 
			return err;	

		if (ctrlc()) {
			printf("Interrupted by user.\n");
			return -1;
		}
		
		Attempts--;

		if (Attempts == 0 ) {
			printf("Timed out.\n");
			return -1;
		}

		if (res[0]  <= 0 || res[0] >3 ) {
			printf("Timed out.\n");
			return -1;
		}

		udelay(1000);
		
	}
	while(res[0] != 0x2);

	memset(&cmd, 0, sizeof(cmd));

	/* APDU read response */
	cmd.cmdidx = 34;
	cmd.cmdarg = 1;
	cmd.resp_type = MMC_RSP_R1;

	data.dest = (char *)res;
	data.blocks = 1;
	data.blocksize = 512;
	data.flags = MMC_DATA_READ;

	err = mmc_send_cmd(mmc, &cmd, &data);
#ifdef PRINT_ASSD
	mmmc_trace_ASSD(mmc, &cmd,err);
#endif
	if (err != 0)
		return err;	

#ifdef PRINT_ASSD
	printf("send CMD13 after CMD 34\n");
#endif	
	
	CMD13Res = CMD13() ;

	if  (CMD13Res < 0 || CMD13Res !=0x0900)
	{
			printf("ASSD command unsuccessfull");
			err = -2;
			
	}	
	
#ifdef PRINT_ASSD
	printf("APDU Response\n");
	print_buffer(0x0000000, res, 1, 512, 0);
#endif

	memcpy(APDU, res, 512);		
	return CMD_RET_SUCCESS;
}

static int do_dp_dongle(cmd_tbl_t *cmdtp, int flag,
	int argc, char * const argv[])
{
	const int ret = doDongle();

	if (CMD_RET_SUCCESS == ret)
	{
		printf("OK.\n");
	}
	else
	{
		printf("dp dongle failed.\n");
	}
	return ret;
}

static int do_dp_login(cmd_tbl_t *cmdtp, int flag,
		       int argc, char * const argv[])
{
	int ret = CMD_RET_SUCCESS;
	int state = 0;
	STATUS_RESPONSE response;
	STATUS_RESPONSE responseCCID;

	if (argc != 1)
	{
		return CMD_RET_USAGE;
	}

	printf("\n Swissbit Data Protection Login (Protocol Version 1)\n");

	ret = dp_status(&response);
		      
	if (CMD_RET_SUCCESS != ret)
	{
	 	printf("No Swissbit Data Protection card found in SD card slot!\n");
		printf(
	"Please insert an activated Swissbit Data Protection card and reboot.\n");
		goto failed;
	}

	/* Check CCID                                                               */
	initCCID();
	/* After initCCID we have CCID pointers ready, check these					*/

	if (NULL != udevCCID)
	{
		printf("Found a Swissbit CCID Device\n");
		/* Switch to CCID                                                       */
		setTarget(DEVICE_CCID);

		ret = dp_status(&responseCCID);
		if (CMD_RET_SUCCESS == ret)
		{
			if (0 == responseCCID.state)
			{
				printf("Your Swissbit CCID Device is not activated\n");
			}
			else if (1 == responseCCID.state)
			{
				printf("Your Swissbit CCID Device is already unlocked!\n");
			}
			else if (2 == responseCCID.state)
			{
				printf("Your Swissbit CCID Device is locked!\n");
				printf("Unlocking Swissbit CCID Device...\n");
				const bool isCCIDHashed = (response.flags & 0x10);
				readAndExecutePolicy(responseCCID.state, isCCIDHashed);
			}
		}
		/* Back to ASSD                                                         */
		setTarget(DEVICE_ASSD);
	}
	else
	{
		printf("No Swissbit CCID Device found\n");
	}

	state = response.state;
	printf("State is %d \n", state);

	if (0 == state)
	{
		// All activation should be done with the card manager
printf("Your Swissbit Data Protection card has not been activated yet!\n");
		printf("Please activate the card with cardManager and reboot!\n");
		goto failed;
#if 0
		ret = activateAndVerify(&state);
		if (ret != CMD_RET_SUCCESS)
		{
			printf("Failed to activateAndVerify!\n");
			ret = CMD_RET_FAILURE;
			goto failed;
		}
#endif
	}
	if (1 == state)
	{
		printf("Your Swissbit Data Protection card is already unlocked!\n");
		printf("You can boot now.\n");
		goto failed;
	}

	const bool isHashed = (response.flags & 0x10);
	ret = readAndExecutePolicy(state, isHashed);
failed:
	return ret;
}



static cmd_tbl_t cmd_dp[] = 
{
	U_BOOT_CMD_MKENT(dev, 2, 0, do_dp_dev, "", ""),
	U_BOOT_CMD_MKENT(cmd, 2, 0, do_dp_cmd, "", ""),
	U_BOOT_CMD_MKENT(status, 2, 0, do_dp_status, "", ""),
	U_BOOT_CMD_MKENT(nvramstatus, 2, 0, do_dp_nvramstatus, "", ""),
	U_BOOT_CMD_MKENT(configurenvram, 6, 0, do_dp_configureNVRAM, "", ""),
	U_BOOT_CMD_MKENT(configsize, 2, 0, do_dp_sizeConfig, "", ""),
	U_BOOT_CMD_MKENT(configread, 4, 0, do_dp_readConfig, "", ""),
	U_BOOT_CMD_MKENT(configwrite, 3, 0, do_dp_writeConfig, "", ""),
	U_BOOT_CMD_MKENT(activate, 4, 0, do_dp_activate, "", ""),
	U_BOOT_CMD_MKENT(deactivate, 2, 0, do_dp_deactivate, "", ""),
	U_BOOT_CMD_MKENT(verify, 2, 0, do_dp_verify, "", ""),
	U_BOOT_CMD_MKENT(hashed, 3, 0, do_dp_hashed, "", ""),
	U_BOOT_CMD_MKENT(extended, 2, 0, do_dp_extended, "", ""),
	U_BOOT_CMD_MKENT(id, 2, 0, do_dp_id, "", ""),
	U_BOOT_CMD_MKENT(cdrom, 2, 0, do_dp_cdrom, "", ""),
	U_BOOT_CMD_MKENT(exception, 3, 0, do_dp_exception, "", ""),
	U_BOOT_CMD_MKENT(usb, 3, 0, do_dp_usb, "", ""),
	U_BOOT_CMD_MKENT(login, 1, 0, do_dp_login, "", ""),
	U_BOOT_CMD_MKENT(play, 1, 0, do_dp_play, "", ""),
	U_BOOT_CMD_MKENT(test, 2, 0, do_dp_test, "", ""),
	U_BOOT_CMD_MKENT(dongle, 1, 0, do_dp_dongle, "", "")
};

static int do_dpops(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	cmd_tbl_t *cp;

	cp = find_cmd_tbl(argv[1], cmd_dp, ARRAY_SIZE(cmd_dp));

	/* Drop the dp command */
	argc--;
	argv++;

	if (cp == NULL || argc > cp->maxargs)
	{
		return CMD_RET_USAGE;
	}
	if (flag == CMD_FLAG_REPEAT && !cp->repeatable)
		return CMD_RET_SUCCESS;

	return cp->cmd(cmdtp, flag, argc, argv);
}

U_BOOT_CMD(
	dp, 6, 0, do_dpops,
	"SwissBit Data Protection storage device",
	"\n"
	"  dev [<dev>] - get or set current MMC device ID\n"
	"  status [<address>] - get security status\n"
	"  activate <password> <SO password> <retry counter> - activate security\n"
	"  deactivate <SO password> - deactivate security\n"
	"  verify <password> - perform PIN entry\n"
	"  hashed PIN|SO-PIN <password> - perform hashed PIN entry\n"
	"  id [<address>]- get unique card ID\n"
	"  test - execute test sequence\n"
	"  extended <flags> - set extended security flags\n"
	"  cdrom <state> - set CD-ROM protection\n"
	"  exception <write> <read> - set write/read exception address\n"
	"  cmd <apdu> - send raw C-APDU and print R-APDU\n"
	);
